/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client;

import java.util.HashSet;

import com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;

/**
 * <p>
 * Abstract base class for {@link ChannelAccessMonitor} implementations. This
 * class takes care of handling the monitor listeners.
 * </p>
 * 
 * <p>
 * This class uses a {@link HashSet} to keep all
 * {@link ChannelAccessMonitorListener ChannelAccessMonitorListeners} that are
 * registered with this monitor. Derived classes should call the
 * {@link #notifyListeners(ChannelAccessGettableValue)} and
 * {@link #notifyListeners(ChannelAccessStatus, String)} methods in order to
 * notify the listeners. Implementations may choose to override the
 * {@link #notifyListener(ChannelAccessMonitorListener, ChannelAccessGettableValue)}
 * and
 * {@link #notifyListener(ChannelAccessMonitorListener, ChannelAccessStatus, String)}
 * methods in order to customize the handling of notifications (e.g. catch
 * certain exceptions possibly thrown by listeners).
 * </p>
 * 
 * @author Sebastian Marsching
 *
 * @param <ValueType>
 *            The exact type of the Channel Access values that are provided by
 *            the monitor events.
 */
public abstract class AbstractChannelAccessMonitor<ValueType extends ChannelAccessGettableValue<?>>
        implements ChannelAccessMonitor<ValueType> {

    private String lastEventErrorMessage;
    private ChannelAccessStatus lastEventErrorStatus;
    private ValueType lastEventValue;
    private boolean lastEventWasAnError;
    private HashSet<ChannelAccessMonitorListener<? super ValueType>> listeners = new HashSet<ChannelAccessMonitorListener<? super ValueType>>();

    /**
     * Synchronization object that this class synchronizes on when modifying the
     * set of registered listeners or triggering the listeners from
     * {@link #notifyListeners(ChannelAccessGettableValue)} or
     * {@link #notifyListener(ChannelAccessMonitorListener, ChannelAccessStatus, String)}
     * .
     */
    protected final Object listenerLock = new Object();

    /**
     * Default constructor. This constructor is supposed to be used by classes
     * that inherit from this class. It initializes the monitor with an empty
     * set of listeners.
     */
    protected AbstractChannelAccessMonitor() {
    }

    /**
     * Notifies the listeners about an error. This method delegates calling each
     * listener to
     * {@link #notifyListener(ChannelAccessMonitorListener, ChannelAccessStatus, String)}
     * so that derived classes can take extra precautions (e.g. catch certain
     * exceptions thrown by a listener). This method also saves the last error
     * so that listeners that are registered after the event can be notified
     * with the error status.
     * 
     * @param status
     *            Channel Access status code associated with the error.
     * @param message
     *            text message describing the error in greater detail (might be
     *            <code>null</code>).
     */
    protected void notifyListeners(ChannelAccessStatus status, String message) {
        synchronized (listenerLock) {
            lastEventWasAnError = true;
            lastEventErrorStatus = status;
            lastEventErrorMessage = message;
            // It certainly is not optimal to call the listeners while we hold
            // the lock. However, if we did not, we could not guarantee that a
            // listener is not notified after having been removed.
            for (ChannelAccessMonitorListener<? super ValueType> listener : listeners) {
                notifyListener(listener, status, message);
            }
        }
    }

    /**
     * Notifies the listeners about a new value that was received. This method
     * delegates calling each listener to
     * {@link #notifyListener(ChannelAccessMonitorListener, ChannelAccessGettableValue)}
     * so that derived classes can take extra precautions (e.g. catch certain
     * exceptions thrown by a listener). This method also saves the last value
     * so that listeners that are registered after the event can be notified
     * with the latest value.
     * 
     * @param value
     *            value that has been received from the server.
     */
    protected void notifyListeners(ValueType value) {
        synchronized (listenerLock) {
            lastEventWasAnError = false;
            lastEventValue = value;
            // It certainly is not optimal to call the listeners while we hold
            // the lock. However, if we did not, we could not guarantee that a
            // listener is not notified after having been removed.
            for (ChannelAccessMonitorListener<? super ValueType> listener : listeners) {
                notifyListener(listener, value);
            }
        }
    }

    /**
     * Notifies a listener about an error. This method can be overridden in
     * order to modify how the listener is called. For example, some
     * implementations might want to catch certain exceptions thrown by a
     * listener.
     * 
     * @param listener
     *            the listener that should be notified. Notifications should
     *            happen by calling the listener's
     *            {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, ChannelAccessStatus, String)
     *            monitorError(...)} method.
     * @param status
     *            Channel Access status code associated with the error.
     * @param message
     *            text message describing the error in greater detail (might be
     *            <code>null</code>).
     */
    protected void notifyListener(
            ChannelAccessMonitorListener<? super ValueType> listener,
            ChannelAccessStatus status, String message) {
        listener.monitorError(this, status, message);
    }

    /**
     * Notifies a listener about a monitor event with a value that was received
     * from the server. This method can be overridden in order to modify how the
     * listener is called. For example, some implementations might want to catch
     * certain exceptions thrown by a listener.
     * 
     * @param listener
     *            the listener that should be notified. Notifications should
     *            happen by calling the listener's
     *            {@link ChannelAccessMonitorListener#monitorEvent(ChannelAccessMonitor, ChannelAccessGettableValue)
     *            monitorEvent(...)} method.
     * @param value
     *            value that has been received from the server.
     */
    protected void notifyListener(
            ChannelAccessMonitorListener<? super ValueType> listener,
            ValueType value) {
        listener.monitorEvent(this, value);
    }

    @Override
    public boolean addMonitorListener(
            ChannelAccessMonitorListener<? super ValueType> listener) {
        if (listener == null) {
            throw new NullPointerException(
                    "The listener parameter must not be null.");
        }
        synchronized (listenerLock) {
            boolean notRegisteredYet = listeners.add(listener);
            if (notRegisteredYet) {
                // Calling the listener while holding the lock is not perfect,
                // but it is the only way to ensure that the listener is not
                // called twice with the same state.
                // We notify the listener with the last value that we received
                // (if we have one) and with the error status, if the last
                // notification was an error status. This gives the listener a
                // chance to use the last good value, even if there was an error
                // in the meantime.
                if (lastEventValue != null) {
                    notifyListener(listener, lastEventValue);
                }
                if (lastEventWasAnError) {
                    notifyListener(listener, lastEventErrorStatus,
                            lastEventErrorMessage);
                }
            }
            return notRegisteredYet;
        }
    }

    @Override
    public boolean removeMonitorListener(
            ChannelAccessMonitorListener<? super ValueType> listener) {
        synchronized (listenerLock) {
            return listeners.remove(listener);
        }
    }

}
