/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client;

import com.aquenos.epics.jackie.client.beacon.DefaultBeaconDetector;
import com.aquenos.epics.jackie.client.internal.ChannelAccessChannelImpl;
import com.aquenos.epics.jackie.client.internal.ChannelPool;
import com.aquenos.epics.jackie.client.internal.ServerPool;
import com.aquenos.epics.jackie.client.resolver.DefaultChannelNameResolver;
import com.aquenos.epics.jackie.common.exception.ErrorHandler;

/**
 * <p>
 * Default implementation of a {@link ChannelAccessClient}. This implementation
 * is configurable and should thus be usable for virtually any application.
 * </p>
 * 
 * <p>
 * This implementation chooses to keep weak references to its channels, the
 * monitors, and even the client itself. This means that if all references are
 * lost, a monitor, a channel, or the whole client will eventually get garbage
 * collected and destroyed. However, as the garbage collector is
 * non-deterministic, you should still use explicit resource management where
 * feasible. However, if your resource management does not correctly cover all
 * cases, the garbage collector should eventually take care of freeing
 * resources, unless you keep strong references to unused objects.
 * </p>
 * 
 * <p>
 * This implementation notifies listeners within the thread that triggers the
 * event. Therefore, listeners should only perform quick operations and they
 * must not block. Care must be taken when listeners acquire a mutex because
 * typically, the code calling the listener will hold a mutex that protects its
 * list of listeners and possibly other attributes. If the listener acquires a
 * mutex, care must be taken that this mutex is never acquired when calling any
 * method of the object that notifies the listener. All exceptions thrown by
 * listeners are caught and forwarded to the configured {@link ErrorHandler}.
 * </p>
 * 
 * <p>
 * This client is configurable through its
 * {@link ChannelAccessClientConfiguration} and its
 * {@link ClientThreadingStrategy}. By default, it uses a default
 * {@link ChannelAccessClientConfiguration} and an instance of
 * {@link DefaultClientThreadingStrategy}.
 * </p>
 * 
 * <p>
 * Typically, each {@link ClientThreadingStrategy} can only be used with a
 * single client, so a new instance should be created for each client that is
 * created.
 * </p>
 * 
 * <p>
 * This client implementation creates instances of {@link DefaultBeaconDetector}
 * and {@link DefaultChannelNameResolver} for detecting beacon anomalies and
 * resolving channel names.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class DefaultChannelAccessClient implements ChannelAccessClient {

    private ChannelAccessClientConfiguration configuration;
    private ClientThreadingStrategy threadingStrategy;

    private DefaultBeaconDetector beaconDetector;
    private DefaultChannelNameResolver resolver;
    private ServerPool serverPool;
    private ChannelPool channelPool;

    /**
     * Creates a Channel Access client using default options. Uses the
     * {@link ChannelAccessClientConfiguration#ChannelAccessClientConfiguration()}
     * constructor to create a configuration and the
     * {@link DefaultClientThreadingStrategy#DefaultClientThreadingStrategy(ErrorHandler)}
     * constructor (passing the {@link ErrorHandler} from the configuration) to
     * create its configuration and threading strategy.
     */
    public DefaultChannelAccessClient() {
        this(null, null);
    }

    /**
     * Creates a Channel Access client using the specified configuration. The
     * client's threading strategy is created using @link
     * DefaultClientThreadingStrategy
     * #DefaultClientThreadingStrategy(ErrorHandler)} constructor (passing the
     * {@link ErrorHandler} from the configuration).
     * 
     * @param configuration
     *            client configuration. If <code>null</code>, a
     *            default-constructed {@link ChannelAccessClientConfiguration}
     *            is used.
     */
    public DefaultChannelAccessClient(
            ChannelAccessClientConfiguration configuration) {
        this(configuration, null);
    }

    /**
     * Creates a Channel Access client using the specified threading strategy.
     * The configuration is created using the
     * {@link ChannelAccessClientConfiguration#ChannelAccessClientConfiguration()}
     * constructor.
     * 
     * @param threadingStrategy
     *            threading strategy to be used by this client. Typically,
     *            threading strategies are not reusable, meaning that a separate
     *            instance must be created for each client. If <code>null</code>
     *            , a {@link DefaultClientThreadingStrategy} is created.
     */
    public DefaultChannelAccessClient(ClientThreadingStrategy threadingStrategy) {
        this(null, threadingStrategy);
    }

    /**
     * Creates a Channel Access client using the specified configuration and
     * threading strategy.
     * 
     * @param configuration
     *            client configuration. If <code>null</code>, a
     *            default-constructed {@link ChannelAccessClientConfiguration}
     *            is used.
     * @param threadingStrategy
     *            threading strategy to be used by this client. Typically,
     *            threading strategies are not reusable, meaning that a separate
     *            instance must be created for each client. If <code>null</code>
     *            , a {@link DefaultClientThreadingStrategy} is created.
     */
    public DefaultChannelAccessClient(
            ChannelAccessClientConfiguration configuration,
            ClientThreadingStrategy threadingStrategy) {
        if (configuration == null) {
            configuration = new ChannelAccessClientConfiguration();
        }
        if (threadingStrategy == null) {
            threadingStrategy = new DefaultClientThreadingStrategy(
                    configuration.getErrorHandler());
        }
        this.configuration = configuration;
        this.threadingStrategy = threadingStrategy;
        this.beaconDetector = new DefaultBeaconDetector(
                configuration.getBeaconDetectorConfiguration());
        this.resolver = new DefaultChannelNameResolver(
                configuration.getResolverConfiguration(), beaconDetector);
        this.threadingStrategy.setBeaconDetector(this.beaconDetector);
        this.threadingStrategy.setChannelNameResolver(this.resolver);
        this.serverPool = new ServerPool(this.beaconDetector,
                this.configuration, this.threadingStrategy);
        this.channelPool = new ChannelPool(this.configuration, this.resolver,
                this.serverPool);
    }

    @Override
    public void destroy() {
        // We destroy the threading strategy first. This ensures that all
        // internal threads are stopped and we can destroy the other components
        // without interfering with these threads.
        threadingStrategy.destroy();
        // Destroying the resolver before destroying the server pool (and thus
        // the servers) should ensure that no new channels are created while
        // destroying the servers. Creating a channel should fail because it
        // cannot add a search request to the resolver.
        resolver.destroy();
        beaconDetector.destroy();
        // Destroy the server pool will destroy all servers which should also
        // destroy all channels. Therefore, we do not have to destroy the
        // channel pool explicitly.
        serverPool.destroy();
    }

    @Override
    public ChannelAccessChannel getChannel(String channelName) {
        // Trying to create a new channel when this client has already been
        // destroyed will result in an IllegalStateException because the channel
        // cannot be registered with the resolver. This is exactly what we want.
        return new ChannelAccessChannelImpl(channelName, this, channelPool,
                threadingStrategy);
    }

    @Override
    public ChannelAccessClientConfiguration getConfiguration() {
        return configuration;
    }

    /**
     * <p>
     * Finalizes this object, destroying the client. This method calls
     * {@link #destroy()} followed by <code>super.finalize()</code>. This
     * ensures that a client is destroyed when it is not used any longer.
     * However, the point in time at which <code>finalize()</code> is called is
     * non-deterministic. For this reason, the <code>destroy()</code> method
     * should be called explicitly whenever this is feasible.
     * </p>
     * 
     * <p>
     * From the documentation of <code>Object.finalize()</code>:
     * </p>
     * 
     * <p>
     * Called by the garbage collector on an object when garbage collection
     * determines that there are no more references to the object. A subclass
     * overrides the <code>finalize</code> method to dispose of system resources
     * or to perform other cleanup.
     * </p>
     * 
     * <p>
     * The general contract of <tt>finalize</tt> is that it is invoked if and
     * when the Java™ virtual machine has determined that there is no longer any
     * means by which this object can be accessed by any thread that has not yet
     * died, except as a result of an action taken by the finalization of some
     * other object or class which is ready to be finalized. The
     * <tt>finalize</tt> method may take any action, including making this
     * object available again to other threads; the usual purpose of
     * <tt>finalize</tt>, however, is to perform cleanup actions before the
     * object is irrevocably discarded. For example, the finalize method for an
     * object that represents an input/output connection might perform explicit
     * I/O transactions to break the connection before the object is permanently
     * discarded.
     * </p>
     * 
     * <p>
     * The Java programming language does not guarantee which thread will invoke
     * the <tt>finalize</tt> method for any given object. It is guaranteed,
     * however, that the thread that invokes finalize will not be holding any
     * user-visible synchronization locks when finalize is invoked. If an
     * uncaught exception is thrown by the finalize method, the exception is
     * ignored and finalization of that object terminates.
     * </p>
     * 
     * <p>
     * After the <tt>finalize</tt> method has been invoked for an object, no
     * further action is taken until the Java virtual machine has again
     * determined that there is no longer any means by which this object can be
     * accessed by any thread that has not yet died, including possible actions
     * by other objects or classes which are ready to be finalized, at which
     * point the object may be discarded.
     * </p>
     * 
     * <p>
     * The <tt>finalize</tt> method is never invoked more than once by a Java
     * virtual machine for any given object.
     * </p>
     * 
     * <p>
     * Any exception thrown by the <code>finalize</code> method causes the
     * finalization of this object to be halted, but is otherwise ignored.
     *
     * @throws Throwable
     *             the <code>Exception</code> raised by this method
     */
    @Override
    protected void finalize() throws Throwable {
        // If there is no (strong) reference to this client, it means that it is
        // not used any longer and it can be stopped. The channels hold a strong
        // reference to the client, so if any channel is still in use, the
        // client is not going to be destroyed.
        try {
            destroy();
        } finally {
            super.finalize();
        }
    }

}
