/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.io.internal;

import java.net.InetSocketAddress;

import com.aquenos.epics.jackie.client.beacon.BeaconDetector;
import com.aquenos.epics.jackie.client.beacon.BeaconDetector.BeaconListener;
import com.aquenos.epics.jackie.common.io.CommunicationController;
import com.aquenos.epics.jackie.common.io.TimerProcessor;

/**
 * <p>
 * Watchdog for a connection with a Channel Access server. The watchdog listens
 * for beacons from the server and, if needed, periodically sends echo messages
 * to the server in order to check whether the connection is working.
 * </p>
 * 
 * <p>
 * With the exception of the {@link #destroy()} method, the watchdog is not safe
 * for use by multiple threads and all of its methods should only be called from
 * the communication controller's thread.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class ConnectionEchoWatchdog {

    /**
     * Interface for sending echo messages. This is the interface through which
     * the watchdog tells its connection to send an echo message.
     * 
     * @author Sebastian Marsching
     */
    public interface EchoMessageSender {

        /**
         * Sends an echo message to the server.
         */
        void sendEchoMessage();

    }

    /**
     * Listener for connection responsiveness changes. This is the interface
     * through which the watchdog sends notifications when it detects that the
     * connection has become unresponsive or that it has become responsive
     * again.
     * 
     * @author Sebastian Marsching
     */
    public interface WatchdogListener {

        /**
         * Notifies the listener that the watchdog detected a change in the
         * connection's responsiveness. When the watchdog is started, it always
         * assumes that the connection is responsive. This means that the first
         * event will always be a notification about an unresponsive connection.
         * 
         * @param responsive
         *            <code>true</code> if the connection has been unresponsive
         *            before and is now responsive, <code>false</code> if the
         *            connection has been responsive before and is now
         *            unresponsive.
         */
        void connectionStateChanged(boolean responsive);

    }

    private static final long ECHO_INTERVAL = 30000L;
    private static final long ECHO_MAX_RESPONSE_TIME = 5000L;

    // We start with the beaconAnomalyDetected flag set to true, because we do
    // not know whether there have been irregular beacons before the watchdog
    // was created.
    private boolean beaconAnomalyDetected = true;
    private BeaconDetector beaconDetector;
    private BeaconListener beaconListener = new BeaconListener() {
        @Override
        public void beaconReceived(InetSocketAddress serverAddress,
                final boolean regularBeacon) {
            // The beacon listener can be (and most likely will be) called from
            // a different thread. Therefore, we schedule a timer for execution
            // as soon as possible, so that we can process the beacon in the
            // correct thread.
            communicationController.registerTimer(new TimerProcessor() {
                @Override
                public void processTimer() {
                    ConnectionEchoWatchdog.this.beaconReceived(regularBeacon);
                }
            }, 0L);
        }
    };
    private CommunicationController communicationController;
    private volatile boolean destroyed = false;
    private boolean echoResponsePending = false;
    private boolean echoTimeoutDetected = false;
    private TimerProcessor echoTimer = new TimerProcessor() {
        @Override
        public void processTimer() {
            echoTimerExpired();
        }
    };
    private WatchdogListener listener;
    private EchoMessageSender messageSender;
    private long nextEchoResponseTime = 0L;
    private long nextEchoTime = 0L;
    private boolean responsive = true;
    private InetSocketAddress serverAddress;

    /**
     * Creates a watchdog. The watchdog is not started automatically. It should
     * be started after the connection has been established by calling the
     * {@link #start()} method.
     * 
     * @param beaconDetector
     *            beacon detector to be used for receiving beacons from the
     *            server. If <code>null</code>, beacons are not used for
     *            assessing the connection health.
     * @param communicationController
     *            communication controller to use for registering timers (never
     *            <code>null</code>). Typically, this should be the same
     *            communication controller that is also used by the connection.
     * @param echoMessageSender
     *            interface through which echo messages are sent (never
     *            <code>null</code>). This interface is typically implemented by
     *            the connection.
     * @param listener
     *            listener that is notified when this watchdog detects a change
     *            of the connection responsiveness.
     * @param serverAddress
     *            IP address and port of the server (never <code>null</code>).
     *            This information is used when registering a listener with the
     *            beacon detector.
     */
    public ConnectionEchoWatchdog(BeaconDetector beaconDetector,
            CommunicationController communicationController,
            EchoMessageSender echoMessageSender, WatchdogListener listener,
            InetSocketAddress serverAddress) {
        this.beaconDetector = beaconDetector;
        this.communicationController = communicationController;
        this.messageSender = echoMessageSender;
        this.listener = listener;
        this.serverAddress = serverAddress;
    }

    private void beaconReceived(boolean regularBeacon) {
        // We ignore beacons when this watchdog has been destroyed.
        if (destroyed) {
            return;
        }
        // This method is executed in the correct thread (the communication
        // thread). Therefore we do not have to worry about race conditions.
        // A regular beacon is an indicator for a good connection to the server.
        // Therefore, we can postpone sending the next echo request.
        if (regularBeacon && !echoResponsePending && !beaconAnomalyDetected) {
            scheduleEchoRequest();
        } else if (!regularBeacon) {
            beaconAnomalyDetected = true;
        }

    }

    private void echoTimerExpired() {
        // If the watchdog has already been destroyed, we do not want to
        // schedule any more timers.
        if (destroyed) {
            return;
        }
        long currentTime = System.currentTimeMillis();
        // Are we waiting for an echo response or do we have to send a new echo
        // message? We only run this block if a timeout was not already
        // detected. If a timeout was detected, this flag will be reset the next
        // time an echo request is sent.
        if (echoResponsePending && nextEchoResponseTime <= currentTime
                && !echoTimeoutDetected) {
            echoTimeoutDetected = true;
            if (responsive) {
                responsive = false;
                listener.connectionStateChanged(false);
            }
        } else if (!echoResponsePending && nextEchoTime <= currentTime) {
            sendEchoRequest();
        }
    }

    private void scheduleEchoRequest() {
        // If the watchdog has already been destroyed, we do not want to
        // schedule any more timers.
        if (destroyed) {
            return;
        }
        this.nextEchoTime = System.currentTimeMillis() + ECHO_INTERVAL;
        communicationController.registerTimer(echoTimer, this.nextEchoTime);
    }

    private void scheduleEchoResponseCheck() {
        // If the watchdog has already been destroyed, we do not want to
        // schedule any more timers.
        if (destroyed) {
            return;
        }
        this.nextEchoResponseTime = System.currentTimeMillis()
                + ECHO_MAX_RESPONSE_TIME;
        communicationController.registerTimer(echoTimer,
                this.nextEchoResponseTime);
    }

    private void sendEchoRequest() {
        // If the watchdog has already been destroyed, we do not want to
        // schedule any more timers.
        if (destroyed) {
            return;
        }
        echoResponsePending = true;
        echoTimeoutDetected = false;
        messageSender.sendEchoMessage();
        scheduleEchoResponseCheck();
    }

    /**
     * Called by the connection when an echo message is received. If an echo
     * response was pending, the echo timer is reset. If a response was pending
     * but it arrived to late, another echo request is sent.
     */
    public void echoMessageReceived() {
        // We only act on an echo response that we expected.
        if (echoResponsePending) {
            // If the echo response is late, we send another request. If it is
            // in time, it means that our connection is working and thus we can
            // schedule the next check in the future and notify listeners that
            // the connection is in a good condition.
            if (echoTimeoutDetected) {
                sendEchoRequest();
            } else {
                echoResponsePending = false;
                scheduleEchoRequest();
                if (!responsive) {
                    responsive = true;
                    listener.connectionStateChanged(true);
                }
            }
        }
    }

    /**
     * Called by the connection when any message arrives. An arriving message is
     * typically an indicator for a good connection and thus sending the next
     * echo request is delayed unless there is a pending echo request. For an
     * echo message, this method should be called after calling
     * {@link #messageReceived()}.
     */
    public void messageReceived() {
        // A received message is an indicator for a good connection. Therefore,
        // we can reset the beacon anomaly state and postpone the next echo
        // request. However, if an echo response is pending, we still wait for
        // that response.
        if (!echoResponsePending) {
            beaconAnomalyDetected = false;
            scheduleEchoRequest();
        }
    }

    /**
     * Called by the connection when a congestion of the receive buffers is
     * detected. This delays the next echo request or the timeout for a pending
     * echo response.
     */
    public void receiveCongestionDetected() {
        // If we are receiving a lot of data we want to postpone the check for
        // a pending echo response or the next echo request. If we are receiving
        // data, this indicates a good connection and an echo response might get
        // stuck because of all the other data anyway.
        if (echoResponsePending) {
            scheduleEchoResponseCheck();
        } else {
            scheduleEchoRequest();
        }
    }

    /**
     * Called by the connection when a congestion of the send buffers is
     * detected and thus a request that has been sent by the watchdog might not
     * have been actually sent to the server yet. This delays the timeout for
     * the pending echo response.
     */
    public void sendCongestionDetected() {
        // If a lot of data to be sent has accumulated, the echo request might
        // still be stuck thus we want to delay the check for the response. As
        // long as data can be sent, the connection is probably fine anyway.
        if (echoResponsePending) {
            scheduleEchoResponseCheck();
        }
    }

    /**
     * Starts this watchdog. This registers this watchdog with the beacon
     * detector and schedules the first echo request to be sent.
     */
    public void start() {
        // If the watchdog has already been destroyed, we do not want to
        // start anything.
        if (destroyed) {
            return;
        }
        if (beaconDetector != null) {
            beaconDetector.addBeaconListener(serverAddress, beaconListener);
        }
        scheduleEchoRequest();
    }

    /**
     * Marks this watchdog as destroyed. This has the consequence that it
     * deleted its registration with the beacon detector and does not schedule
     * any more timers. This method should always be called when the connection
     * assoicated with this watchdog is destroyed in order to avoid resource
     * leaks.
     */
    public void destroy() {
        destroyed = true;
        if (beaconDetector != null) {
            beaconDetector.removeBeaconListener(serverAddress, beaconListener);
        }
    }

}
