/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.resolver.internal;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.DatagramChannel;
import java.nio.channels.SelectionKey;
import java.util.Collection;
import java.util.LinkedList;

/**
 * <p>
 * Asynchronous UDP sender. Queues packets and send them to multiple target
 * addresses.
 * </p>
 * 
 * <p>
 * This class is only intended for use by {@link UdpResolver}. It is closely
 * tied to that class and has only been separated into a class of its own to
 * keep {@link UdpResolver} smaller.
 * </p>
 * 
 * <p>
 * This class is not thread safe, but the instantiating {@link UdpResolver}
 * ensures that it is only used in a safe way.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class UdpSender {

    private ByteBuffer currentBuffer = null;
    private LinkedList<ByteBuffer> freeBuffers = new LinkedList<ByteBuffer>();
    private LinkedList<ByteBuffer> pendingBuffers = new LinkedList<ByteBuffer>();
    private LinkedList<InetSocketAddress> pendingTargetAddresses = new LinkedList<InetSocketAddress>();
    private LinkedList<InetSocketAddress> targetAddresses = new LinkedList<InetSocketAddress>();
    private DatagramChannel channel;
    private SelectionKey selectionKey;

    /**
     * Creates a UDP sender that sends packets using the specified channel. The
     * channel must be non-blocking.
     * 
     * @param channel
     *            non-blocking channel which is used to send UDP packets.
     */
    public UdpSender(DatagramChannel channel) {
        assert (channel != null);
        assert (channel.isBlocking() == false);
        this.channel = channel;
    }

    /**
     * Returns a buffer from the internal free list. Returns <code>null</code>
     * if the internal free list is empty. The buffers on the free list are
     * buffers that have been passed to this class for sending and have been
     * processed completely. They have already been cleared, so they are
     * immediately usable for writing data to them.
     * 
     * @return buffer from the free list or <code>null</code> if the free list
     *         is empty.
     */
    public ByteBuffer nextFreeBuffer() {
        return freeBuffers.poll();
    }

    /**
     * Tells whether there are buffers in this sender's queue. A buffer is kept
     * in the queue until it has been sent to all target addresses.
     * 
     * @return <code>true</code> if there are buffers which are waiting to be
     *         sent, <code>false</code> if the queue is empty.
     */
    public boolean pendingWrite() {
        return (currentBuffer != null || !pendingBuffers.isEmpty());
    }

    /**
     * Tries to send the pending buffers using the channel. Returns when all
     * data has been sent or the channel does not accept more data. Buffers that
     * have been sent completely are added to the free list. If this method
     * throws an exception, this class will not try again to send the same
     * buffer to the same target address but will continue with the next target
     * address (or buffer if the error occurred for the last target address).
     * This method must not be called before
     * {@link #setSelectionKey(SelectionKey)} has been called.
     * 
     * @throws ClosedChannelException
     *             if the channel's
     *             {@link DatagramChannel#send(ByteBuffer, java.net.SocketAddress)
     *             send} method throws such an exception.
     * @throws IOException
     *             if the channel's
     *             {@link DatagramChannel#send(ByteBuffer, java.net.SocketAddress)
     *             send} method throws such an exception.
     */
    public void processWrite() throws ClosedChannelException, IOException {
        if (currentBuffer == null) {
            currentBuffer = pendingBuffers.poll();
            pendingTargetAddresses.addAll(targetAddresses);
        }
        while (currentBuffer != null) {
            while (!pendingTargetAddresses.isEmpty()) {
                InetSocketAddress targetAddress = pendingTargetAddresses.poll();
                if (channel.send(currentBuffer, targetAddress) == 0) {
                    // The socket's send buffer is full, so we cannot send a
                    // packet at the moment. We try the same target address
                    // later.
                    pendingTargetAddresses.addFirst(targetAddress);
                    return;
                } else {
                    // The packet has been sent, so we have to rewind the buffer
                    // for the next send operation.
                    currentBuffer.rewind();
                }
            }
            // Recycle the buffer.
            currentBuffer.clear();
            freeBuffers.add(currentBuffer);
            // Get the next buffer.
            currentBuffer = pendingBuffers.poll();
            if (currentBuffer != null) {
                pendingTargetAddresses.addAll(targetAddresses);
            }
        }
        // There are no pending buffers left, thus we are not longer interested
        // whether the channel is ready for writes.
        disableWriteEvent();
    }

    /**
     * Adds the specified buffer to the list of buffers to be send. The buffer
     * must be small enough to fit into a single UDP packet. Otherwise, the
     * {@link #processWrite()} method will stop processing any data because it
     * will be stuck with this buffer. This method must not be called before
     * {@link #setSelectionKey(SelectionKey)} has been called.
     * 
     * @param buffer
     *            buffer to be sent as a UDP packet.
     */
    public void sendBuffer(ByteBuffer buffer) {
        assert (buffer != null);
        pendingBuffers.add(buffer);
        enableWriteEvent();
    }

    /**
     * Sets the selection key for the channel. The selection key is used to
     * register this sender's interest in being informed when more data can be
     * written. Must be called before trying to send data.
     * 
     * @param selectionKey
     *            selection key for the channel.
     */
    public void setSelectionKey(SelectionKey selectionKey) {
        assert (selectionKey != null);
        this.selectionKey = selectionKey;
    }

    /**
     * Sets the list of target addresses that shall receive UDP packets. This
     * method may be called from time to time to update the target addresses.
     * Updating the addresses affects buffers in the queue but not the buffer
     * that is currently being processed.
     * 
     * @param targetAddresses
     *            list of target addresses for UDP packets.
     */
    public void setTargetAddresses(
            Collection<? extends InetSocketAddress> targetAddresses) {
        this.targetAddresses.clear();
        this.targetAddresses.addAll(targetAddresses);
    }

    private void enableWriteEvent() {
        selectionKey.interestOps(selectionKey.interestOps()
                | SelectionKey.OP_WRITE);
    }

    private void disableWriteEvent() {
        selectionKey.interestOps(selectionKey.interestOps()
                & ~SelectionKey.OP_WRITE);
    }

}
