/*
 * Copyright 2014-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.util.concurrent.ConcurrentHashMap;

import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.exception.ChannelAccessException;

/**
 * <p>
 * Channel Access status codes. The status codes are used to signal exceptional
 * conditions over the network. The {@link #forStatusCode(int)} method can be
 * used to convert a status code that has been sent over the network as an
 * integer.
 * </p>
 * <p>
 * A status code has a message number and a severity. For each message, there is
 * a severity, that should be associated with this message. For status codes
 * received from remote peers however, there is no guarantee that the expected
 * severity is associated with a certain message. Therefore, status codes should
 * only be compared regarding the message, not the severity. The easiest way to
 * do this is using the {@link #matchesMessage(StatusMessage)} method.
 * </p>
 * <p>
 * Unlike the EPICS C-library, this library does not return status codes to
 * signal error conditions internally. Instead, Java exceptions are preferred
 * for this purpose. An error condition that is associated with a network error
 * (e.g. that has been received from a peer) is typically signaled by a
 * {@link ChannelAccessException} that wraps the status code.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessStatus {

    /**
     * Message describing a type of error. The message is taken from a limited
     * set of descriptions for error causes.
     * 
     * @author Sebastian Marsching
     */
    public static enum StatusMessage {
        /**
         * Normal successful completion. This message is associated with a
         * severity of {@link StatusSeverityLevel#SUCCESS}.
         */
        ECA_NORMAL(StatusSeverityLevel.SUCCESS, 0,
                "Normal successful completion"),

        /**
         * Maximum simultaneous IOC connections exceeded. This message is
         * associated with a severity of {@link StatusSeverityLevel#ERROR}. This
         * status code is deprecated and should not be used by new servers.
         * However, it still might be sent by old servers.
         */
        @Deprecated
        ECA_MAXIOC(StatusSeverityLevel.ERROR, 1,
                "Maximum simultaneous IOC connections exceeded"),

        /**
         * Unknown Internet host. This message is associated with a severity of
         * {@link StatusSeverityLevel#ERROR}. This status code is deprecated and
         * should not be used by new servers. However, it still might be sent by
         * old servers.
         */
        @Deprecated
        ECA_UKNHOST(StatusSeverityLevel.ERROR, 2, "Unknown Internet host"),

        /**
         * Unknown Internet service. This message is associated with a severity
         * of {@link StatusSeverityLevel#ERROR}. This status code is deprecated
         * and should not be used by new servers. However, it still might be
         * sent by old servers.
         */
        @Deprecated
        ECA_UKNSERV(StatusSeverityLevel.ERROR, 3, "Unknown Internet service"),

        /**
         * Unable to allocate a new socket. This message is associated with a
         * severity of {@link StatusSeverityLevel#ERROR}. This status code is
         * deprecated and should not be used by new servers. However, it still
         * might be sent by old servers.
         */
        @Deprecated
        ECA_SOCK(StatusSeverityLevel.ERROR, 4,
                "Unable to allocate a new socket"),

        /**
         * Unable to connect to Internet host or service. This message is
         * associated with a severity of {@link StatusSeverityLevel#WARNING}.
         * This status code is deprecated and should not be used by new servers.
         * However, it still might be sent by old servers.
         */
        @Deprecated
        ECA_CONN(StatusSeverityLevel.WARNING, 5,
                "Unable to connect to Internet host or service"),

        /**
         * Unable to allocate additional dynamic memory. This message is
         * associated with a severity of {@link StatusSeverityLevel#WARNING}.
         */
        ECA_ALLOCMEM(StatusSeverityLevel.WARNING, 6,
                "Unable to allocate additional dynamic memory"),

        /**
         * Unknown IO channel. This message is associated with a severity of
         * {@link StatusSeverityLevel#WARNING}. This status code is deprecated
         * and should not be used by new servers. However, it still might be
         * sent by old servers.
         */
        @Deprecated
        ECA_UKNCHAN(StatusSeverityLevel.WARNING, 7, "Unknown IO channel"),

        /**
         * Record field specified inappropriate for channel specified. This
         * message is associated with a severity of
         * {@link StatusSeverityLevel#WARNING}. This status code is deprecated
         * and should not be used by new servers. However, it still might be
         * sent by old servers.
         */
        @Deprecated
        ECA_UKNFIELD(StatusSeverityLevel.WARNING, 8,
                "Record field specified inappropriate for channel specified"),

        /**
         * The requested data transfer is greater than available memory or
         * EPICS_CA_MAX_ARRAY_BYTES. This message is associated with a severity
         * of {@link StatusSeverityLevel#WARNING}.
         */
        ECA_TOLARGE(
                StatusSeverityLevel.WARNING,
                9,
                "The requested data transfer is greater than available memory or EPICS_CA_MAX_ARRAY_BYTES"),

        /**
         * User specified timeout on IO operation expired. This message is
         * associated with a severity of {@link StatusSeverityLevel#WARNING}.
         */
        ECA_TIMEOUT(StatusSeverityLevel.WARNING, 10,
                "User specified timeout on IO operation expired"),

        /**
         * Sorry, that feature is planned but not supported at this time. This
         * message is associated with a severity of
         * {@link StatusSeverityLevel#WARNING}. This status code is deprecated
         * and should not be used by new servers. However, it still might be
         * sent by old servers.
         */
        @Deprecated
        ECA_NOSUPPORT(StatusSeverityLevel.WARNING, 11,
                "Sorry, that feature is planned but not supported at this time"),

        /**
         * The supplied string is unusually large. This message is associated
         * with a severity of {@link StatusSeverityLevel#WARNING}. This status
         * code is deprecated and should not be used by new servers. However, it
         * still might be sent by old servers.
         */
        @Deprecated
        ECA_STRTOBIG(StatusSeverityLevel.WARNING, 12,
                "The supplied string is unusually large"),

        /**
         * The request was ignored because the specified channel is
         * disconnected. This message is associated with a severity of
         * {@link StatusSeverityLevel#ERROR} . This status code is deprecated
         * and should not be used by new servers. However, it still might be
         * sent by old servers.
         */
        @Deprecated
        ECA_DISCONNCHID(StatusSeverityLevel.ERROR, 13,
                "The request was ignored because the specified channel is disconnected"),

        /**
         * The data type specified is invalid. This message is associated with a
         * severity of {@link StatusSeverityLevel#ERROR}.
         */
        ECA_BADTYPE(StatusSeverityLevel.ERROR, 14,
                "The data type specifed is invalid"),

        /**
         * Remote Channel not found. This message is associated with a severity
         * of {@link StatusSeverityLevel#INFO}. This status code is deprecated
         * and should not be used by new servers. However, it still might be
         * sent by old servers.
         */
        @Deprecated
        ECA_CHIDNOTFND(StatusSeverityLevel.INFO, 15, "Remote Channel not found"),

        /**
         * Unable to locate all user specified channels. This message is
         * associated with a severity of {@link StatusSeverityLevel#INFO}. This
         * status code is deprecated and should not be used by new servers.
         * However, it still might be sent by old servers.
         */
        @Deprecated
        ECA_CHIDRETRY(StatusSeverityLevel.INFO, 16,
                "Unable to locate all user specified channels"),

        /**
         * Channel Access Internal Failure. This message is associated with a
         * fatal severity ({@link StatusSeverityLevel#ERROR} and severe flag
         * set).
         */
        ECA_INTERNAL(StatusSeverityLevel.ERROR, true, 17,
                "Channel Access Internal Failure"),

        /**
         * The requested local DB operation failed. This message is associated
         * with a severity of {@link StatusSeverityLevel#WARNING}. This status
         * code is deprecated and should not be used by new servers. However, it
         * still might be sent by old servers.
         */
        @Deprecated
        ECA_DBLCLFAIL(StatusSeverityLevel.WARNING, 18,
                "The requested local DB operation failed"),

        /**
         * Channel read request failed. This message is associated with a
         * severity of {@link StatusSeverityLevel#WARNING}.
         */
        ECA_GETFAIL(StatusSeverityLevel.WARNING, 19,
                "Channel read request failed"),

        /**
         * Channel write request failed. This message is associated with a
         * severity of {@link StatusSeverityLevel#WARNING}.
         */
        ECA_PUTFAIL(StatusSeverityLevel.WARNING, 20,
                "Channel write request failed"),

        /**
         * Channel subscription request failed. This message is associated with
         * a severity of {@link StatusSeverityLevel#WARNING}. This status code
         * is deprecated and should not be used by new servers. However, it
         * still might be sent by old servers.
         */
        @Deprecated
        ECA_ADDFAIL(StatusSeverityLevel.WARNING, 21,
                "Channel subscription request failed"),

        /**
         * Invalid element count requested. This message is associated with a
         * severity of {@link StatusSeverityLevel#WARNING}.
         */
        ECA_BADCOUNT(StatusSeverityLevel.WARNING, 22,
                "Invalid element count requested"),

        /**
         * Invalid string. This message is associated with a severity of
         * {@link StatusSeverityLevel#ERROR}.
         */
        ECA_BADSTR(StatusSeverityLevel.ERROR, 23, "Invalid string"),

        /**
         * Virtual circuit disconnect. This message is associated with a
         * severity of {@link StatusSeverityLevel#WARNING}.
         */
        ECA_DISCONN(StatusSeverityLevel.WARNING, 24,
                "Virtual circuit disconnect"),

        /**
         * Identical process variable names on multiple servers. This message is
         * associated with a severity of {@link StatusSeverityLevel#WARNING}.
         */
        ECA_DBLCHNL(StatusSeverityLevel.WARNING, 25,
                "Identical process variable names on multiple servers"),

        /**
         * Request inappropriate within subscription (monitor) update callback.
         * This message is associated with a severity of
         * {@link StatusSeverityLevel#ERROR}.
         */
        ECA_EVDISALLOW(StatusSeverityLevel.ERROR, 26,
                "Request inappropriate within subscription (monitor) update callback"),

        /**
         * Database value get for that channel failed during channel search.
         * This message is associated with a severity of
         * {@link StatusSeverityLevel#WARNING}. This status code is deprecated
         * and should not be used by new servers. However, it still might be
         * sent by old servers.
         */
        @Deprecated
        ECA_BUILDGET(StatusSeverityLevel.WARNING, 27,
                "Database value get for that channel failed during channel search"),

        /**
         * Unable to initialize without the vxWorks VX_FP_TASK task option set.
         * This message is associated with a severity of
         * {@link StatusSeverityLevel#WARNING}. This status code is deprecated
         * and should not be used by new servers. However, it still might be
         * sent by old servers.
         */
        @Deprecated
        ECA_NEEDSFP(StatusSeverityLevel.WARNING, 28,
                "Unable to initialize without the vxWorks VX_FP_TASK task option set"),

        /**
         * Event queue overflow has prevented first pass event after event add.
         * This message is associated with a severity of
         * {@link StatusSeverityLevel#WARNING}. This status code is deprecated
         * and should not be used by new servers. However, it still might be
         * sent by old servers.
         */
        @Deprecated
        ECA_OVEFAIL(StatusSeverityLevel.WARNING, 29,
                "Event queue overflow has prevented first pass event after event add"),

        /**
         * Bad event subscription (monitor) identifier. This message is
         * associated with a severity of {@link StatusSeverityLevel#ERROR}.
         */
        ECA_BADMONID(StatusSeverityLevel.ERROR, 30,
                "Bad event subscription (monitor) identifier"),

        /**
         * Remote channel has new network address. This message is associated
         * with a severity of {@link StatusSeverityLevel#WARNING}. This status
         * code is deprecated and should not be used by new servers. However, it
         * still might be sent by old servers.
         */
        @Deprecated
        ECA_NEWADDR(StatusSeverityLevel.WARNING, 31,
                "Remote channel has new network address"),

        /**
         * New or resumed network connection. This message is associated with a
         * severity of {@link StatusSeverityLevel#INFO}. This status code is
         * deprecated and should not be used by new servers. However, it still
         * might be sent by old servers.
         */
        @Deprecated
        ECA_NEWCONN(StatusSeverityLevel.INFO, 32,
                "New or resumed network connection"),

        /**
         * Specified task isn't a member of a CA context. This message is
         * associated with a severity of {@link StatusSeverityLevel#WARNING}.
         * This status code is deprecated and should not be used by new servers.
         * However, it still might be sent by old servers.
         */
        @Deprecated
        ECA_NOCACTX(StatusSeverityLevel.WARNING, 33,
                "Specified task isn't a member of a CA context"),

        /**
         * Attempt to use defunct CA feature failed. This message is associated
         * with a fatal severity ({@link StatusSeverityLevel#ERROR} and severe
         * flag set). This status code is deprecated and should not be used by
         * new servers. However, it still might be sent by old servers.
         */
        @Deprecated
        ECA_DEFUNCT(StatusSeverityLevel.ERROR, true, 34,
                "Attempt to use defunct CA feature failed"),

        /**
         * The supplied string is empty. This message is associated with a
         * severity of {@link StatusSeverityLevel#WARNING}. This status code is
         * deprecated and should not be used by new servers. However, it still
         * might be sent by old servers.
         */
        @Deprecated
        ECA_EMPTYSTR(StatusSeverityLevel.WARNING, 35,
                "The supplied string is empty"),

        /**
         * Unable to spawn the CA repeater thread- auto reconnect will fail.
         * This message is associated with a severity of
         * {@link StatusSeverityLevel#WARNING}. This status code is deprecated
         * and should not be used by new servers. However, it still might be
         * sent by old servers.
         */
        @Deprecated
        ECA_NOREPEATER(StatusSeverityLevel.WARNING, 36,
                "Unable to spawn the CA repeater thread- auto reconnect will fail"),

        /**
         * No channel id match for search reply- search reply ignored. This
         * message is associated with a severity of
         * {@link StatusSeverityLevel#WARNING}. This status code is deprecated
         * and should not be used by new servers. However, it still might be
         * sent by old servers.
         */
        @Deprecated
        ECA_NOCHANMSG(StatusSeverityLevel.WARNING, 37,
                "No channel id match for search reply- search reply ignored"),

        /**
         * Reseting dead connection- will try to reconnect. This message is
         * associated with a severity of {@link StatusSeverityLevel#WARNING}.
         * This status code is deprecated and should not be used by new servers.
         * However, it still might be sent by old servers.
         */
        @Deprecated
        ECA_DLCKREST(StatusSeverityLevel.WARNING, 38,
                "Reseting dead connection- will try to reconnect"),

        /**
         * Server (IOC) has fallen behind or is not responding- still waiting.
         * This message is associated with a severity of
         * {@link StatusSeverityLevel#WARNING}. This status code is deprecated
         * and should not be used by new servers. However, it still might be
         * sent by old servers.
         */
        @Deprecated
        ECA_SERVBEHIND(StatusSeverityLevel.WARNING, 39,
                "Server (IOC) has fallen behind or is not responding- still waiting"),

        /**
         * No Internet interface with broadcast available. This message is
         * associated with a severity of {@link StatusSeverityLevel#WARNING}.
         * This status code is deprecated and should not be used by new servers.
         * However, it still might be sent by old servers.
         */
        @Deprecated
        ECA_NOCAST(StatusSeverityLevel.WARNING, 40,
                "No Internet interface with broadcast available"),

        /**
         * Invalid event selection mask. This message is associated with a
         * severity of {@link StatusSeverityLevel#ERROR}.
         */
        ECA_BADMASK(StatusSeverityLevel.ERROR, 41,
                "Invalid event selection mask"),

        /**
         * IO operations have completed. This message is associated with a
         * severity of {@link StatusSeverityLevel#INFO}.
         */
        ECA_IODONE(StatusSeverityLevel.INFO, 42, "IO operations have completed"),

        /**
         * IO operations are in progress. This message is associated with a
         * severity of {@link StatusSeverityLevel#INFO}.
         */
        ECA_IOINPROGRESS(StatusSeverityLevel.INFO, 43,
                "IO operations are in progress"),

        /**
         * Invalid synchronous group identifier. This message is associated with
         * a severity of {@link StatusSeverityLevel#ERROR}.
         */
        ECA_BADSYNCGRP(StatusSeverityLevel.ERROR, 44,
                "Invalid synchronous group identifier"),

        /**
         * Put callback timed out. This message is associated with a severity of
         * {@link StatusSeverityLevel#ERROR}.
         */
        ECA_PUTCBINPROG(StatusSeverityLevel.ERROR, 45, "Put callback timed out"),

        /**
         * Read access denied. This message is associated with a severity of
         * {@link StatusSeverityLevel#WARNING}.
         */
        ECA_NORDACCESS(StatusSeverityLevel.WARNING, 46, "Read access denied"),

        /**
         * Write access denied. This message is associated with a severity of
         * {@link StatusSeverityLevel#WARNING}.
         */
        ECA_NOWTACCESS(StatusSeverityLevel.WARNING, 47, "Write access denied"),

        /**
         * Requested feature is no longer supported. This message is associated
         * with a severity of {@link StatusSeverityLevel#ERROR}.
         */
        ECA_ANACHRONISM(StatusSeverityLevel.ERROR, 48,
                "Requested feature is no longer supported"),

        /**
         * Empty PV search address list. This message is associated with a
         * severity of {@link StatusSeverityLevel#WARNING}.
         */
        ECA_NOSEARCHADDR(StatusSeverityLevel.WARNING, 49,
                "Empty PV search address list"),

        /**
         * No reasonable data conversion between client and server types. This
         * message is associated with a severity of
         * {@link StatusSeverityLevel#WARNING}.
         */
        ECA_NOCONVERT(StatusSeverityLevel.WARNING, 50,
                "No reasonable data conversion between client and server types"),

        /**
         * Invalid channel identifier. This message is associated with a
         * severity of {@link StatusSeverityLevel#ERROR}.
         */
        ECA_BADCHID(StatusSeverityLevel.ERROR, 51, "Invalid channel identifier"),

        /**
         * Invalid function pointer. This message is associated with a severity
         * of {@link StatusSeverityLevel#ERROR}.
         */
        ECA_BADFUNCPTR(StatusSeverityLevel.ERROR, 52,
                "Invalid function pointer"),

        /**
         * Thread is already attached to a client context. This message is
         * associated with a severity of {@link StatusSeverityLevel#WARNING}.
         */
        ECA_ISATTACHED(StatusSeverityLevel.WARNING, 53,
                "Thread is already attached to a client context"),

        /**
         * Not supported by attached service. This message is associated with a
         * severity of {@link StatusSeverityLevel#WARNING}.
         */
        ECA_UNAVAILINSERV(StatusSeverityLevel.WARNING, 54,
                "Not supported by attached service"),

        /**
         * User destroyed channel. This message is associated with a severity of
         * {@link StatusSeverityLevel#WARNING}.
         */
        ECA_CHANDESTROY(StatusSeverityLevel.WARNING, 55,
                "User destroyed channel"),

        /**
         * Invalid channel priority. This message is associated with a severity
         * of {@link StatusSeverityLevel#ERROR}.
         */
        ECA_BADPRIORITY(StatusSeverityLevel.ERROR, 56,
                "Invalid channel priority"),

        /**
         * Preemptive callback not enabled - additional threads may not join
         * context. This message is associated with a severity of
         * {@link StatusSeverityLevel#ERROR}.
         */
        ECA_NOTTHREADED(StatusSeverityLevel.ERROR, 57,
                "Preemptive callback not enabled - additional threads may not join context"),

        /**
         * Client's protocol revision does not support transfers exceeding 16k
         * bytes. This message is associated with a severity of
         * {@link StatusSeverityLevel#WARNING}.
         */
        ECA_16KARRAYCLIENT(StatusSeverityLevel.WARNING, 58,
                "Client's protocol revision does not support transfers exceeding 16k bytes"),

        /**
         * Virtual circuit connection sequence aborted. This message is
         * associated with a severity of {@link StatusSeverityLevel#WARNING}.
         */
        ECA_CONNSEQTMO(StatusSeverityLevel.WARNING, 59,
                "Virtual circuit connection sequence aborted"),

        /**
         * Virtual circuit unresponsive. This message is associated with a
         * severity of {@link StatusSeverityLevel#WARNING}.
         */
        ECA_UNRESPTMO(StatusSeverityLevel.WARNING, 60,
                "Virtual circuit unresponsive");

        /**
         * All status enum values sorted by their message number. This array is
         * used for quickly retrieving the enum object associated with a certain
         * status code.
         * 
         * @see #forStatusCode(int)
         */
        private final static StatusMessage[] MESSAGE_INSTANCES;

        static {
            StatusMessage[] enumInstances = StatusMessage.values();
            MESSAGE_INSTANCES = new StatusMessage[enumInstances.length];
            for (StatusMessage status : enumInstances) {
                int messageNumber = status.messageNumber;
                if (MESSAGE_INSTANCES[messageNumber] != null) {
                    throw new RuntimeException(
                            "Detected duplicate message number "
                                    + messageNumber + ".");
                }
                MESSAGE_INSTANCES[messageNumber] = status;
            }
        }

        private final StatusSeverity defaultSeverity;
        private final int messageNumber;
        private final String message;
        private final int defaultStatusCode;

        private StatusMessage(StatusSeverityLevel level, int messageNumber,
                String message) {
            this(level, false, messageNumber, message);
        }

        private StatusMessage(StatusSeverityLevel level, boolean severe,
                int messageNumber, String message) {
            this.messageNumber = messageNumber;
            this.message = message;
            this.defaultSeverity = StatusSeverity.SEVERITY_INSTANCES[level
                    .getSeverityNumber() + (severe ? 4 : 0)];
            this.defaultStatusCode = calculateStatusCode(
                    level.getSeverityNumber() + (severe ? 4 : 0), messageNumber);
        }

        private static int calculateStatusCode(int severityNumber,
                int messageNumber) {
            int statusCode = severityNumber & 0x7;
            statusCode |= ((messageNumber) << 3) & 0xfff8;
            return statusCode;
        }

        /**
         * <p>
         * Returns the severity associated with this status code by default.
         * </p>
         * <p>
         * This is the severity that should be used when sending a status
         * containing this message. Status codes received from remote peers
         * however, might have a different severity for the same message.
         * </p>
         * 
         * @return severity usually associated with this status code (never
         *         <code>null</code>).
         */
        public StatusSeverity getDefaultSeverity() {
            return defaultSeverity;
        }

        /**
         * Returns the status code usually associated with this message. This is
         * the status code that should be sent by a server that wants to send an
         * error with this message. The status code is calculated using the
         * number of this message and the severity returned by
         * {@link #getDefaultSeverity()}.
         * 
         * @return standard status code for this message.
         */
        public int getDefaultStatusCode() {
            return defaultStatusCode;
        }

        /**
         * Returns the status code for this message, assuming the severity
         * returned by {@link #getDefaultSeverity()}.
         * 
         * @return status code for this message.
         */
        public ChannelAccessStatus toStatus() {
            return ChannelAccessStatus.forStatusCode(defaultStatusCode);
        }

        /**
         * Returns the textual description associated with this status message.
         * This text is suitable for displaying an error to the user.
         * 
         * @return description associated with this status message.
         */
        @Override
        public String toString() {
            return message;
        }

    }

    /**
     * <p>
     * Severity levels associated with status codes. The severity level
     * indicates into which category a status code belongs. Usually, the
     * behavior of a client does not depend on the severity level associated
     * with a status code.
     * </p>
     * <p>
     * The EPICS C-library knows a severity level <code>FATAL</code>.
     * Technically, this is just the <code>ERROR</code> level combined with the
     * severe flag. In this implementation, the severe flag is kept separately
     * in the {@link StatusSeverity}, so that we can use an enum for the
     * severity level.
     * </p>
     * 
     * @author Sebastian Marsching
     * 
     */
    public static enum StatusSeverityLevel {
        /**
         * Reserved for the {@link StatusMessage#ECA_NORMAL} message.
         */
        SUCCESS(1),

        /**
         * Severity associated with messages that have an informational
         * character and do not necessarily signal an error condition.
         */
        INFO(3),

        /**
         * Severity associated with messages that signal an error condition that
         * is might be transient. Usually this means that an operation failed,
         * however the problem might resolve itself and it might make sense to
         * retry later.
         */
        WARNING(0),
        /**
         * Severity associated with messages that signal an error condition that
         * is caused by a non-transient problem. This means that an operation
         * failed due to a permanent problem (e.g. an invalid parameter value
         * sent by the client).
         */
        ERROR(2);

        private final int severityNumber;

        private StatusSeverityLevel(int severityNumber) {
            this.severityNumber = severityNumber;
        }

        private int getSeverityNumber() {
            return severityNumber;
        }

    }

    /**
     * Severity for status codes. The severity indicates to which category a
     * status code belongs and whether it represents a severe problem.
     * 
     * @author Sebastian Marsching
     */
    public static class StatusSeverity {

        /**
         * All eight variants of the severity class. As there are only four
         * severity levels and one boolean flag in the severity, there cannot be
         * more than these eight severities. We create them once and reuse them
         * for both better performance and simpler comparison. The order in
         * which they are defined in this area is well-chosen, so that it
         * exactly matches the the octal number transferred over the network.
         * Therefore the number from the network can be used as an index into
         * this array to get the right severity.
         */
        private final static StatusSeverity[] SEVERITY_INSTANCES = new StatusSeverity[] {
                new StatusSeverity(StatusSeverityLevel.WARNING, false,
                        564685104),
                new StatusSeverity(StatusSeverityLevel.SUCCESS, false,
                        -1837262005),
                new StatusSeverity(StatusSeverityLevel.ERROR, false,
                        -1706982248),
                new StatusSeverity(StatusSeverityLevel.INFO, false, -413279776),
                new StatusSeverity(StatusSeverityLevel.WARNING, true, 625469114),
                new StatusSeverity(StatusSeverityLevel.SUCCESS, true,
                        -105788039),
                new StatusSeverity(StatusSeverityLevel.ERROR, true, 1424402941),
                new StatusSeverity(StatusSeverityLevel.INFO, true, -1194002941) };

        private int hashCode;
        private final StatusSeverityLevel level;
        private final boolean severe;

        private StatusSeverity(StatusSeverityLevel level, boolean severe,
                int hashCode) {
            this.hashCode = hashCode;
            this.level = level;
            this.severe = severe;
        }

        /**
         * Returns the level of this severity. The level defines to which
         * category a message belongs: success indicator, informational message,
         * warning or error.
         * 
         * @return level associated with this severity.
         */
        public StatusSeverityLevel getLevel() {
            return level;
        }

        /**
         * Returns <code>true</code> if this severity indicates a severe
         * problem. Usually, the severe flag is only used in combination with
         * the {@link StatusSeverityLevel#ERROR} level and is reserved for
         * problems that indicate an unexpected problem in the server.
         * 
         * @return flag indicating a severe problem.
         */
        public boolean isSevere() {
            return severe;
        }

        /**
         * <p>
         * Returns a hash code value for this severity. This method is supported
         * for the benefit of hash tables. As instances of this type are
         * immutable, the hash code is a fixed number that is already determined
         * during construction.
         * </p>
         * 
         * <p>
         * If two severities are equal (as indicated by the return value of
         * {@link #equals(Object)}), they also have the same hash code. However,
         * the fact that two messages have the same hash code does not imply
         * that they are equal, even though the implementation of this method
         * has been carefully designed to make such a collision extremely
         * unlikely.
         * </p>
         * 
         * @return hash code suitable for use by a hash table.
         */
        @Override
        public int hashCode() {
            // We use a fixed hash code because we have a limited set of objects
            // anyway. The numbers that we use are random number so that they
            // should be nicely distributed.
            // Even though we actually have distinct hash codes for all possible
            // values, we do not advertise this fact in the API documentation so
            // that we can change it in the future if it should become
            // necessary.
            return hashCode;
        }

        /**
         * <p>
         * Tells whether the specified object represents the same severity.
         * </p>
         * 
         * <p>
         * Two severities are considered equal if they have the same severity
         * level and if they both have the same "severe" flag (either set for
         * both or not set for both).
         * </p>
         * 
         * @param obj
         *            reference object with which this object is compared.
         * @return <code>true</code> if <code>obj</code> is a severity that has
         *         the same level and severe flag as this severity,
         *         <code>false</code> otherwise.
         */
        @Override
        public boolean equals(Object obj) {
            // Instances of this type are unique, so we can simply check for
            // object equality.
            return this == obj;
        }

        /**
         * Returns a string representation of this severity. The string
         * representation is the string representation of the severity level
         * plus the string "severe" added in parentheses if the severe flag is
         * set. The <code>ERROR</code> severity level is an exception to this
         * rule: If the "severe" bit is set for this level, the string "FATAL"
         * is returned instead of "ERROR".
         * 
         * @return string representation of this severity.
         */
        @Override
        public String toString() {
            if (!severe) {
                return level.name();
            } else {
                switch (level) {
                case ERROR:
                    return "FATAL";
                case INFO:
                    return "INFO (severe)";
                case SUCCESS:
                    return "SUCCESS (severe)";
                case WARNING:
                    return "WARNING (severe)";
                default:
                    throw new RuntimeException(
                            "Unexpected value for enum of type "
                                    + StatusSeverityLevel.class.getName()
                                    + ": " + level.name());
                }
            }
        }

    }

    private static ConcurrentHashMap<Integer, ChannelAccessStatus> generatedStatusMap = new ConcurrentHashMap<Integer, ChannelAccessStatus>();

    private final int statusCode;
    private final StatusSeverity severity;
    private final StatusMessage message;

    private ChannelAccessStatus(int statusCode, StatusSeverity severity,
            StatusMessage message) {
        this.statusCode = statusCode;
        this.severity = severity;
        this.message = message;
    }

    /**
     * Returns the integer representation of this status code. This
     * representation is suitable for sending it over the network. However, it
     * should not be used to compare it with status codes received from a
     * network peer, because the severity might differ. Use the
     * {@link #matchesMessage(StatusMessage)} method instead.
     * 
     * @return integer representation of this status code.
     */
    public int toStatusCode() {
        return statusCode;
    }

    /**
     * Returns the status message associated with this status code. The message
     * identifies the problem and is associated with a textual description.
     * 
     * @return message identifying the status code This might be
     *         <code>null</code> if the message part of the status code
     *         specifies a message number that is not known by this
     *         implementation.
     */
    public StatusMessage getMessage() {
        return message;
    }

    /**
     * <p>
     * Returns the severity associated with this status code.
     * </p>
     * 
     * @return severity associated with this status code (never
     *         <code>null</code>).
     */
    public StatusSeverity getSeverity() {
        return severity;
    }

    /**
     * Returns a text message associated with a certain status code. The
     * severity returned in the resulting string is the actual severity encoded
     * in the status code, not just the severity usually associated with the
     * message number in this library. The return value of this method is the
     * same as <code>forStatusCode(statusCode).toString()</code>.
     * 
     * @param statusCode
     *            integer representation of the status code (e.g. received from
     *            a network peer).
     * @return text message containing the severity and a short description of
     *         the status code (never <code>null</code>, not even for an unknown
     *         status code).
     */
    public static String messageForStatusCode(int statusCode) {
        return forStatusCode(statusCode).toString();
    }

    /**
     * Returns an object representing the specified status code. This object
     * will allow access to the message and severity encoded in the status code.
     * 
     * @param statusCode
     *            integer representation of the status code (e.g. received from
     *            a network peer).
     * @return object representing the status code (never <code>null</code>).
     */
    public static ChannelAccessStatus forStatusCode(int statusCode) {
        statusCode &= 0xffff;
        ChannelAccessStatus status = generatedStatusMap.get(statusCode);
        if (status == null) {
            status = new ChannelAccessStatus(statusCode,
                    severityFromStatusCode(statusCode),
                    messageFromStatusCode(statusCode));
            ChannelAccessStatus existingStatus = generatedStatusMap
                    .putIfAbsent(statusCode, status);
            if (existingStatus != null) {
                return existingStatus;
            }
        }
        return status;
    }

    private static StatusMessage messageFromStatusCode(int statusCode) {
        int messageNumber = (statusCode & 0xfff8) >>> 3;
        if (messageNumber < StatusMessage.MESSAGE_INSTANCES.length) {
            return StatusMessage.MESSAGE_INSTANCES[messageNumber];
        } else {
            return null;
        }
    }

    private static StatusSeverity severityFromStatusCode(int statusCode) {
        int severityNumber = statusCode & 0x7;
        return StatusSeverity.SEVERITY_INSTANCES[severityNumber];
    }

    /**
     * Returns a text message associated with this status code. The severity
     * returned in the resulting string is the actual severity encoded in the
     * status code, not just the severity usually associated with the message
     * number in this library. The return value of this method is the same as
     * <code>forStatusCode(statusCode).toString()</code>.
     * 
     * @return text message containing the severity and a short description of
     *         the status code (never <code>null</code>, not even for an unknown
     *         status code).
     */
    @Override
    public String toString() {
        if (message != null) {
            return new StringBuilder().append(severity.toString()).append(": ")
                    .append(message.toString()).toString();
        } else {
            return "Status code sent by server is not known by this implementation";
        }
    }

    /**
     * <p>
     * Returns a hash code value for this status object. This method is
     * supported for the benefit of hash tables. The hash code is calculated
     * based on the status code so that a different status code typically
     * results in a different hash code.
     * </p>
     * 
     * <p>
     * If two status objects are equal (as indicated by the return value of
     * {@link #equals(Object)}), they also have the same hash code. However, the
     * fact that two status objects have the same hash code does not imply that
     * they are equal.
     * </p>
     * 
     * @return hash code suitable for use by a hash table.
     */
    @Override
    public int hashCode() {
        // Integrating the severity into the hash code is redundant, because
        // the status code already contains the severity. However, it might give
        // a slightly better hash code, because the severity's hashCode method
        // probably does not return nearly consecutive integers.
        return new HashCodeBuilder().append(statusCode).append(severity)
                .toHashCode();
    }

    /**
     * Tells whether the specified object represents the same status object. Two
     * status objects are considered equal if they have the same status code.
     * 
     * @param obj
     *            reference object with which this object is compared.
     * @return <code>true</code> if <code>obj</code> is a status object and has
     *         the same status code as this object, <code>false</code>
     *         otherwise.
     */
    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (!(obj instanceof ChannelAccessStatus)) {
            return false;
        }
        ChannelAccessStatus other = (ChannelAccessStatus) obj;
        return this.statusCode == other.statusCode;
    }

    /**
     * Checks whether the message part of this status code matches a certain
     * message. This is unlike the {@link #equals(Object)} method, that also
     * checks whether the severities match.
     * 
     * @param message
     *            message to compare this status code's message with.
     * @return <code>true</code> if <code>message</code> is the same message as
     *         represented by this status code, <code>false</code> if it is a
     *         different message.
     */
    public boolean matchesMessage(StatusMessage message) {
        // If we do not have a message, this status code has an unknown message
        // number. Therefore it cannot match any StatusMessage instance.
        if (this.message == null) {
            return false;
        }
        return this.message == message;
    }

}
