/*
 * Copyright 2014-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

/**
 * Channel Access protocol version. The messages supported by the protocol and
 * the meaning of header fields depends on the protocol version.
 * 
 * @author Sebastian Marsching
 */
public enum ChannelAccessVersion implements Comparable<ChannelAccessVersion> {

    /**
     * Channel Access version 4.0 (or unknown minor revision)
     */
    V4_0(0),

    /**
     * Channel Access version 4.1
     */
    V4_1(1),

    /**
     * Channel Access version 4.2
     */
    V4_2(2),

    /**
     * Channel Access version 4.3
     */
    V4_3(3),

    /**
     * Channel Access version 4.4
     */
    V4_4(4),

    /**
     * Channel Access version 4.5
     */
    V4_5(5),

    /**
     * Channel Access version 4.6
     */
    V4_6(6),

    /**
     * Channel Access version 4.7
     */
    V4_7(7),

    /**
     * Channel Access version 4.8
     */
    V4_8(8),

    /**
     * Channel Access version 4.9
     */
    V4_9(9),

    /**
     * Channel Access version 4.10
     */
    V4_10(10),

    /**
     * Channel Access version 4.11
     */
    V4_11(11),

    /**
     * Channel Access version 4.12
     */
    V4_12(12),

    /**
     * Channel Access version 4.13 (or a newer version)
     */
    V4_13(13);

    /**
     * Newest version supported by this library. At the moment, this is Channel
     * Access version 4.13.
     */
    public final static ChannelAccessVersion NEWEST_SUPPORTED_VERSION = V4_13;

    private int majorVersion;
    private int minorVersion;

    private final static ChannelAccessVersion[] MINOR_VERSION_TO_ENUM = new ChannelAccessVersion[] {
            V4_0, V4_1, V4_2, V4_3, V4_4, V4_5, V4_6, V4_7, V4_8, V4_9, V4_10,
            V4_11, V4_12, V4_13 };

    private final static int MAXIMUM_MINOR_VERSION = MINOR_VERSION_TO_ENUM.length - 1;

    private ChannelAccessVersion(int majorVersion, int minorVersion) {
        this.majorVersion = majorVersion;
        this.minorVersion = minorVersion;
    }

    private ChannelAccessVersion(int minorVersion) {
        this(4, minorVersion);
    }

    /**
     * Returns the major protocol version. At the moment, this code only
     * supports Channel Access version 4.x, so this method always returns
     * <code>4</code>.
     * 
     * @return major protocol version.
     */
    public int getMajorVersion() {
        return majorVersion;
    }

    /**
     * Returns the minor protocol version. The minor version defines which
     * message types are supported and the format of messages might vary between
     * versions. In general, a component that understands a newer minor version
     * of the protocol should also understand all messages that might be sent by
     * a component speaking an older minor version. However, the support for
     * vary old versions (older than Channel Access 4.4) might be limited.
     * 
     * @return minor protocol version. This is always a non-negative number.
     */
    public int getMinorVersion() {
        return minorVersion;
    }

    /**
     * Returns the minor protocol version as a short. This method is supplied
     * for convenience, because the Channel Access header stores the minor
     * version as an unsigned 16-bit integer. This method returns the same value
     * as {@link #getMinorVersion()}, however, this method might not work, if
     * the minor version number is greater than or equal to 2^16. As this is not
     * going to happen in the foreseeable future, it should be safe to use this
     * method. If the minor version number is greater than or equal to 2^15 but
     * less than 2^16, the short returned will be a negative number.
     * 
     * @return minor protocol version as a short. This number might be negative,
     *         if the version number is greater than or equal to 2^15.
     * 
     * @throws RuntimeException
     *             if the minor version number is greater than or equal to 2^16.
     */
    public short getMinorVersionAsShort() {
        if (minorVersion >= 65536) {
            throw new RuntimeException("Cannot cast version number "
                    + minorVersion + " to a 16-bit integer.");
        }
        return (short) minorVersion;
    }

    /**
     * Returns whether this protocol version includes another protocol version.
     * This method returns <code>true</code> if this version's minor version is
     * equal to or greater than the minor version of <code>otherVersion</code>
     * and the major versions are the same. It returns <code>false</code> if
     * <code>otherVersion</code> is newer than this version or the major
     * versions do not match.
     * 
     * @param otherVersion
     *            version to compare with this version. Must not be
     *            <code>null</code>.
     * @return <code>true</code> if this version also supports all messages
     *         valid for <code>otherVersion</code>, <code>false</code>
     *         otherwise.
     */
    public boolean includes(ChannelAccessVersion otherVersion) {
        return (otherVersion.minorVersion <= this.minorVersion && otherVersion.majorVersion == this.majorVersion);
    }

    /**
     * Returns the version object for the specified major and minor version. At
     * the moment, only major version 4 is supported. If the specified minor
     * version is newer than the newest known minor version, the newest known
     * minor version is used. This is okay because newer versions are expected
     * to maintain backward compatibility.
     * 
     * @param majorVersion
     *            major version to look for. At the moment only the major
     *            version 4 is supported.
     * @param minorVersion
     *            minor version to look for. Must be non-negative.
     * @return the version object for the specified major and minor version,
     *         never <code>null</code>.
     * @throws IllegalArgumentException
     *             if the major version is not supported or the minor version is
     *             negative.
     */
    public static ChannelAccessVersion forVersion(int majorVersion,
            int minorVersion) {
        if (majorVersion != 4) {
            throw new IllegalArgumentException(
                    "Only major version 4 is supported.");
        }
        return forMinorVersion(minorVersion);
    }

    /**
     * Returns the version object for the specified minor version, assuming
     * major version 4. If the specified minor version is newer than the newest
     * known minor version, the newest known minor version is used. This is okay
     * because newer versions are expected to maintain backward compatibility.
     * 
     * @param minorVersion
     *            minor version to look for. Must be non-negative.
     * @return the version object for the specified major and minor version,
     *         never <code>null</code>.
     * @throws IllegalArgumentException
     *             if the minor version is negative.
     */
    public static ChannelAccessVersion forMinorVersion(int minorVersion) {
        if (minorVersion < 0) {
            throw new IllegalArgumentException(
                    "Minor version must be greater than or equal to 0.");
        }
        if (minorVersion > MAXIMUM_MINOR_VERSION) {
            minorVersion = MAXIMUM_MINOR_VERSION;
        }
        return MINOR_VERSION_TO_ENUM[minorVersion];
    }

    /**
     * Returns the version object for the specified minor version, assuming
     * major version 4. If the specified minor version is newer than the newest
     * known minor version, the newest known minor version is used. This is okay
     * because newer versions are expected to maintain backward compatibility.
     * The passed number is interpreted as an unsigned integer, therefore
     * negative values are mapped to positive values above
     * {@link Short#MAX_VALUE}. This method is provided for convenience, because
     * Channel Access uses an unsigned 16-bit integer for the wire transport.
     * 
     * @param minorVersion
     *            minor version to look for. May be negative.
     * @return the version object for the specified major and minor version,
     *         never <code>null</code>.
     */
    public static ChannelAccessVersion forMinorVersion(short minorVersion) {
        return forMinorVersion(minorVersion & 0xffff);
    }

    /**
     * Returns a string representation of this version number. The string
     * representation has the form "majorVersion.minorVersion".
     * 
     * @return string representation of this version number.
     */
    @Override
    public String toString() {
        return "" + majorVersion + "." + minorVersion;
    }

}
