/*
 * Copyright 2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/**
 * Listenable future implementation that simply delegates all method calls to
 * another listenable future. While this implementation can be used on its own,
 * it is most useful as a base class for implementations want to intercept
 * certain method calls while simply passing other ones through.
 * 
 * @author Sebastian Marsching
 *
 * @param <V>
 *            The result type returned by the future.
 */
public class DelegatingListenableFuture<V> implements ListenableFuture<V> {

    /**
     * Delegate to which all method calls are delegated.
     */
    protected ListenableFuture<? extends V> delegate;

    /**
     * Creates a listenable future that delegates all method calls to the
     * specified listenable future.
     * 
     * @param delegate
     *            listenable future that will receive all method calls that are
     *            made on this listenable future.
     */
    public DelegatingListenableFuture(ListenableFuture<? extends V> delegate) {
        this.delegate = delegate;
    }

    @Override
    public boolean cancel(boolean mayInterruptIfRunning) {
        return delegate.cancel(mayInterruptIfRunning);
    }

    @Override
    public boolean isCancelled() {
        return delegate.isCancelled();
    }

    @Override
    public boolean isDone() {
        return delegate.isDone();
    }

    @Override
    public V get() throws InterruptedException, ExecutionException {
        return delegate.get();
    }

    @Override
    public V get(long timeout, TimeUnit unit) throws InterruptedException,
            ExecutionException, TimeoutException {
        return delegate.get(timeout, unit);
    }

    @Override
    public boolean addCompletionListener(
            FutureCompletionListener<? super V> listener) {
        return delegate.addCompletionListener(listener);
    }

    @Override
    public boolean removeCompletionListener(
            FutureCompletionListener<? super V> listener) {
        return delegate.removeCompletionListener(listener);
    }

}
