/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.util.concurrent.Executor;
import java.util.concurrent.Future;

/**
 * <p>
 * Future that allows asynchronous notifications.
 * </p>
 * 
 * <p>
 * This kind of future can be used to be notified when the result is available,
 * thus removing the need for a blocking thread that waits on the result.
 * </p>
 * 
 * @author Sebastian Marsching
 *
 * @param <V>
 *            The result type returned by the future.
 */
public interface ListenableFuture<V> extends Future<V> {

    /**
     * <p>
     * Registers a completion listener with this future. This listener is
     * notified when the future completes ({@link #isDone()} returns
     * <code>true</code>). If the future completes before the listener is
     * registered, the listener is notified immediately while being added.
     * </p>
     * 
     * <p>
     * If the specified listener is already registered with this future, no
     * action is taken and this method returns <code>false</code>.
     * </p>
     * 
     * <p>
     * <strong>Important note</strong>: Typically, the listener is executed by
     * the thread that sets the future's value. For this reason, it is important
     * that the listener does not block or perform any long-running operations.
     * Instead, the listener should delegate any such operation to a different
     * thread (e.g. using an {@link Executor}). Any violation of this rule might
     * result in unexpected or unpredictable behavior. In particular, a blocking
     * listener might cause a dead-lock, which results in the whole application
     * being stalled indefinitely.
     * </p>
     * 
     * @param listener
     *            listener to be registered with this future.
     * @return <code>true</code> if the listener has been added,
     *         <code>false</code> if it has already been added earlier.
     */
    boolean addCompletionListener(FutureCompletionListener<? super V> listener);

    /**
     * <p>
     * Removes the registration of a completion listener with this future. After
     * removing, the listener is not going to be notified when the future
     * completes.
     * </p>
     * 
     * <p>
     * If the specified listener is not currently registered with this future,
     * no action is taken and this method returns <code>false</code>.
     * </p>
     * 
     * @param listener
     *            listener to be unregistered from this future.
     * @return <code>true</code> if the listener has been removed,
     *         <code>false</code> if it has already been removed earlier of if
     *         has never been added.
     */
    boolean removeCompletionListener(
            FutureCompletionListener<? super V> listener);

}
