/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

import java.util.concurrent.ConcurrentHashMap;

/**
 * <p>
 * Channel Access alarm status codes. The status codes are used to signal the
 * cause of an alarm over the network. The {@link #forAlarmStatusCode(short)}
 * method can be used to convert a status code that has been sent over the
 * network as a short integer.
 * </p>
 * <p>
 * Some alarm status codes have a predefined name and meaning. This name can be
 * retrieved using the {@link #toString()} method. For an alarm status without a
 * predefined meaning, the string "ALARM_###" is returned, where "###" is the
 * status code (interpreted as an unsigned number).
 * </p>
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessAlarmStatus {

    /**
     * No alarm (everything is OK).
     */
    public static final ChannelAccessAlarmStatus NO_ALARM = new ChannelAccessAlarmStatus(
            "NO_ALARM", (short) 0);

    /**
     * Error while trying to read a value.
     */
    public static final ChannelAccessAlarmStatus READ = new ChannelAccessAlarmStatus(
            "READ", (short) 1);

    /**
     * Error while trying to write a value.
     */
    public static final ChannelAccessAlarmStatus WRITE = new ChannelAccessAlarmStatus(
            "WRITE", (short) 2);

    /**
     * Value is above upper alarm limit.
     */
    public static final ChannelAccessAlarmStatus HIHI = new ChannelAccessAlarmStatus(
            "HIHI", (short) 3);

    /**
     * Value is above upper warning limit.
     */
    public static final ChannelAccessAlarmStatus HIGH = new ChannelAccessAlarmStatus(
            "HIGH", (short) 4);

    /**
     * Value is below lower alarm limit.
     */
    public static final ChannelAccessAlarmStatus LOLO = new ChannelAccessAlarmStatus(
            "LOLO", (short) 5);

    /**
     * Value is below lower warning limit.
     */
    public static final ChannelAccessAlarmStatus LOW = new ChannelAccessAlarmStatus(
            "LOW", (short) 6);

    /**
     * Value represents a state that is an alarm condition.
     */
    public static final ChannelAccessAlarmStatus STATE = new ChannelAccessAlarmStatus(
            "STATE", (short) 7);

    /**
     * Value had a change of state that signals an alarm condition.
     */
    public static final ChannelAccessAlarmStatus COS = new ChannelAccessAlarmStatus(
            "COS", (short) 8);

    /**
     * Device support signaled a communication problem.
     */
    public static final ChannelAccessAlarmStatus COMM = new ChannelAccessAlarmStatus(
            "COMM", (short) 9);

    /**
     * Device support signaled that an asynchronous device timed out.
     */
    public static final ChannelAccessAlarmStatus TIMEOUT = new ChannelAccessAlarmStatus(
            "TIMEOUT", (short) 10);

    /**
     * Device support signaled that a hardware limit alarm has been triggered.
     */
    public static final ChannelAccessAlarmStatus HWLIMIT = new ChannelAccessAlarmStatus(
            "HWLIMIT", (short) 11);

    /**
     * Calculation failed.
     */
    public static final ChannelAccessAlarmStatus CALC = new ChannelAccessAlarmStatus(
            "CALC", (short) 12);

    /**
     * Record could not be processed repeatedly because it is locked.
     */
    public static final ChannelAccessAlarmStatus SCAN = new ChannelAccessAlarmStatus(
            "SCAN", (short) 13);

    /**
     * Record link could not be processed. This might be caused by an invalid
     * link or because the linked record is in an invalid state.
     */
    public static final ChannelAccessAlarmStatus LINK = new ChannelAccessAlarmStatus(
            "LINK", (short) 14);

    /**
     * Record soft alarm. This alarm can be caused by an invalid record
     * configuration.
     */
    public static final ChannelAccessAlarmStatus SOFT = new ChannelAccessAlarmStatus(
            "SOFT", (short) 15);

    /**
     * Record alarm indicating that no subroutine has been set or the configured
     * subroutine is invalid.
     */
    public static final ChannelAccessAlarmStatus BAD_SUB = new ChannelAccessAlarmStatus(
            "BAD_SUB", (short) 16);

    /**
     * Value is undefined. This error is typically caused by a record that has
     * never been processed and thus does not have a well-defined value yet.
     */
    public static final ChannelAccessAlarmStatus UDF = new ChannelAccessAlarmStatus(
            "UDF", (short) 17);

    /**
     * Alarm raised when a record has been disabled and it is configured to
     * raise an alarm in this case.
     */
    public static final ChannelAccessAlarmStatus DISABLE = new ChannelAccessAlarmStatus(
            "DISABLE", (short) 18);

    /**
     * Simulated alarm. This alarm can be triggered by a record in simulation
     * mode, that has been configured to cause an alarm.
     */
    public static final ChannelAccessAlarmStatus SIMM = new ChannelAccessAlarmStatus(
            "SIMM", (short) 19);

    /**
     * Read access error.
     */
    public static final ChannelAccessAlarmStatus READ_ACCESS = new ChannelAccessAlarmStatus(
            "READ_ACCESS", (short) 20);

    /**
     * Write access error.
     */
    public static final ChannelAccessAlarmStatus WRITE_ACCESS = new ChannelAccessAlarmStatus(
            "WRITE_ACCESS", (short) 21);

    private static ConcurrentHashMap<Short, ChannelAccessAlarmStatus> generatedAlarmStatusMap = new ConcurrentHashMap<Short, ChannelAccessAlarmStatus>();

    private final static ChannelAccessAlarmStatus[] statusCodeToStatus = new ChannelAccessAlarmStatus[] {
            NO_ALARM, READ, WRITE, HIHI, HIGH, LOLO, LOW, STATE, COS, COMM,
            TIMEOUT, HWLIMIT, CALC, SCAN, LINK, SOFT, BAD_SUB, UDF, DISABLE,
            SIMM, READ_ACCESS, WRITE_ACCESS };

    private String name;
    private short statusCode;

    private ChannelAccessAlarmStatus(String name, short statusCode) {
        this.name = name;
        this.statusCode = (short) statusCode;
    }

    /**
     * Returns the status code associated with this alarm status. This is an
     * integer that uniquely identifies this alarm status. This number is
     * suitable to be sent over the network.
     * 
     * @return integer number identifying this alarm status.
     * @see #forAlarmStatusCode(short)
     */
    public short toAlarmStatusCode() {
        return statusCode;
    }

    /**
     * Returns a string representation of this alarm-status object. For the
     * well-known status codes, this is the name of the static field in this
     * class that represents the respective status code. For unknown status
     * codes this is the string "ALARM_" concatenated with the string
     * representation of the numeric status code.
     * 
     * @return string representation of this status code.
     */
    @Override
    public String toString() {
        return name;
    }

    /**
     * Returns the alarm status for the given <code>statusCode</code>. This can
     * be used to translate a status code received from the network into a
     * severity.
     * 
     * @param statusCode
     *            integer number identifying the alarm status.
     * @return the alarm status identified by the status code (never
     *         <code>null</code>).
     */
    public static ChannelAccessAlarmStatus forAlarmStatusCode(short statusCode) {
        // We use an array for the predefined codes, because this is the most
        // common case and thus we want to keep it fast.
        if (statusCode >= 0 && statusCode < statusCodeToStatus.length) {
            return statusCodeToStatus[statusCode];
        }

        ChannelAccessAlarmStatus status = generatedAlarmStatusMap
                .get(statusCode);
        if (status == null) {
            int intStatusCode = statusCode;
            if (intStatusCode < 0) {
                intStatusCode += 65536;
            }
            status = new ChannelAccessAlarmStatus("ALARM_" + intStatusCode,
                    statusCode);
            ChannelAccessAlarmStatus existingStatus = generatedAlarmStatusMap
                    .putIfAbsent(statusCode, status);
            if (existingStatus != null) {
                return existingStatus;
            }
        }
        return status;
    }

    /**
     * <p>
     * Returns a hash code value for this alarm-status object. This method is
     * supported for the benefit of hash tables. The hash code is calculated
     * based on the numeric status code of this alarm-status object.
     * </p>
     * 
     * <p>
     * If two alarm-status objects are equal (as indicated by the return value
     * of {@link #equals(Object)}), they also have the same hash code. However,
     * the fact that two alarm-status objects have the same hash code does not
     * imply that they are equal.
     * </p>
     * 
     * @return hash code suitable for use by a hash table.
     */
    @Override
    public int hashCode() {
        return 41 + 19 * statusCode;
    }

    /**
     * Tells whether the specified object represents the same alarm-status
     * object. Two alarm-status objects are considered equal if and only if they
     * have the same status code.
     * 
     * @param obj
     *            reference object with which this object is compared.
     * @return <code>true</code> if <code>obj</code> is an alarm-status object
     *         with the same status code as this object, <code>false</code>
     *         otherwise.
     */
    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (!(obj instanceof ChannelAccessAlarmStatus)) {
            return false;
        }
        return ((ChannelAccessAlarmStatus) obj).statusCode == this.statusCode;
    }

}
