/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value.internal;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;
import java.util.Collection;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyString;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmSeverity;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmString;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyString;
import com.aquenos.epics.jackie.common.value.ChannelAccessString;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeStamp;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeString;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueFactory;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Base class for Channel Access string type. This class contains the various
 * specializations of the string type as inlined classes.
 * 
 * @param <ImplementationType>
 *            type of the actual implementation.
 * 
 * @author Sebastian Marsching
 */
public abstract class ChannelAccessStringBase<ImplementationType extends ChannelAccessStringBase<ImplementationType>>
        extends AbstractStringBase<ImplementationType> implements
        ChannelAccessString {

    /**
     * Implementation of DBR_STRING.
     */
    public static final class ChannelAccessStringImpl extends
            ChannelAccessStringBase<ChannelAccessStringImpl> implements
            ChannelAccessSimpleOnlyString {

        /**
         * Creates a DBR_STRING.
         * 
         * @param value
         *            string value.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        public ChannelAccessStringImpl(Collection<String> value, Charset charset) {
            super(value, charset);
        }

        /**
         * Creates a read-only DBR_STRING. All read operations are delegated to
         * <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessStringImpl(ChannelAccessStringImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessStringImpl(Charset charset) {
            super(charset);
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_STRING;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.size());
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessSimpleOnlyString asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessStringImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            return super.equals(obj);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return super.hashCode();
        }

        @Override
        public ChannelAccessSimpleOnlyString clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessSimpleOnlyString) super.clone();
        }

        /**
         * Creates a DBR_STRING from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @param charset
         *            character-set used for encoding and decoding strings.
         * @return DBR_STRING instance created from the serialized data read
         *         from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessStringImpl deserialize(
                ByteSource byteSource, int count, Charset charset) {
            ChannelAccessStringImpl value = new ChannelAccessStringImpl(charset);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Base for string with alarms.
     */
    public static abstract class ChannelAccessAlarmStringBase<ImplementationType extends ChannelAccessAlarmStringBase<ImplementationType>>
            extends ChannelAccessStringBase<ImplementationType> implements
            ChannelAccessAlarmString {

        /**
         * Alarm status.
         */
        protected ChannelAccessAlarmStatus alarmStatus;

        /**
         * Alarm severity.
         */
        protected ChannelAccessAlarmSeverity alarmSeverity;

        /**
         * Base constructor.
         * 
         * @param value
         *            string value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessAlarmStringBase(Collection<String> value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity, Charset charset) {
            super(value, charset);
            this.alarmStatus = alarmStatus;
            this.alarmSeverity = alarmSeverity;
        }

        /**
         * Base constructor for read-only variant.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessAlarmStringBase(ImplementationType delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessAlarmStringBase(Charset charset) {
            super(charset);
        }

        @Override
        public ChannelAccessAlarmStatus getAlarmStatus() {
            if (delegate != null) {
                return delegate.getAlarmStatus();
            }
            return this.alarmStatus;
        }

        @Override
        public void setAlarmStatus(ChannelAccessAlarmStatus alarmStatus) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            assert (alarmStatus != null);
            this.alarmStatus = alarmStatus;
        }

        @Override
        public ChannelAccessAlarmSeverity getAlarmSeverity() {
            if (delegate != null) {
                return delegate.getAlarmSeverity();
            }
            return this.alarmSeverity;
        }

        @Override
        public void setAlarmSeverity(ChannelAccessAlarmSeverity alarmSeverity) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            assert (alarmSeverity != null);
            this.alarmSeverity = alarmSeverity;
        }

        /**
         * Deserializes the alarm status from the byte source.
         * 
         * @param byteSource
         *            byte source to read from.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        protected void deserializeAlarms(ByteSource byteSource) {
            this.alarmStatus = ValueCodecUtils
                    .deserializeAlarmStatus(byteSource);
            this.alarmSeverity = ValueCodecUtils
                    .deserializeAlarmSeverity(byteSource);
        }

        /**
         * Serializes the alarm status to the byte sink.
         * 
         * @param byteSink
         *            byte sink to write to.
         */
        protected void serializeAlarms(ByteSink byteSink) {
            ValueCodecUtils.serializeAlarms(byteSink, alarmStatus,
                    alarmSeverity);
        }

        @Override
        public boolean equals(Object obj) {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            @SuppressWarnings("unchecked")
            ChannelAccessAlarmStringBase<ImplementationType> other = (ChannelAccessAlarmStringBase<ImplementationType>) obj;
            return new EqualsBuilder()
                    .append(this.alarmStatus, other.alarmStatus)
                    .append(this.alarmSeverity, other.alarmSeverity).isEquals();
        }

        @Override
        public int hashCode() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.alarmStatus).append(this.alarmSeverity)
                    .toHashCode();
        }

        @Override
        public ChannelAccessAlarmString clone() {
            // We do not handle the read-only variant here. If this value is
            // read-only, this method should never be called by the child class.
            assert (!isReadOnly());
            return (ChannelAccessAlarmString) super.clone();
        }

    }

    /**
     * Implementation of DBR_STS_STRING.
     */
    public static final class ChannelAccessAlarmStringImpl extends
            ChannelAccessAlarmStringBase<ChannelAccessAlarmStringImpl>
            implements ChannelAccessAlarmOnlyString {

        /**
         * Actual type that shall be represented.
         */
        protected ChannelAccessValueType type;

        /**
         * Creates a DBR_STS_STRING, DBR_GR_STRING, or DBR_CTRL_STRING.
         * 
         * @param type
         *            actual type that shall be represented.
         * @param value
         *            string value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        public ChannelAccessAlarmStringImpl(ChannelAccessValueType type,
                Collection<String> value, ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity, Charset charset) {
            super(value, alarmStatus, alarmSeverity, charset);
            // This representation should only be use for specific types.
            assert (type == ChannelAccessValueType.DBR_STS_STRING
                    || type == ChannelAccessValueType.DBR_GR_STRING || type == ChannelAccessValueType.DBR_CTRL_STRING);
            this.type = type;
        }

        /**
         * Creates a read-only DBR_STS_STRING, DBR_GR_STRING, or
         * DBR_CTRL_STRING. All read operations are delegated to
         * <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessAlarmStringImpl(
                ChannelAccessAlarmStringImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param type
         *            actual type that shall be represented.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessAlarmStringImpl(ChannelAccessValueType type,
                Charset charset) {
            super(charset);
            // This representation should only be use for specific types.
            assert (type == ChannelAccessValueType.DBR_STS_STRING
                    || type == ChannelAccessValueType.DBR_GR_STRING || type == ChannelAccessValueType.DBR_CTRL_STRING);
            this.type = type;
        }

        @Override
        public ChannelAccessValueType getType() {
            if (delegate != null) {
                return delegate.getType();
            }
            return type;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            if (delegate != null) {
                delegate.serialize(byteSink, count);
                return;
            }
            assert (count >= 0);
            assert (count <= value.size());
            serializeAlarms(byteSink);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessAlarmOnlyString asReadOnlyValue() {
            if (delegate != null) {
                return this;
            } else {
                return new ChannelAccessAlarmStringImpl(this);
            }
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            ChannelAccessAlarmStringImpl other = (ChannelAccessAlarmStringImpl) obj;
            return this.type.equals(other.type);
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.type).toHashCode();
        }

        @Override
        public ChannelAccessAlarmOnlyString clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessAlarmOnlyString) super.clone();
        }

        /**
         * Creates a DBR_STS_STRING, DBR_GR_STRING, or DBR_CTRL_STRING from its
         * serialized representation.
         * 
         * @param type
         *            actual value type.
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @param charset
         *            character-set used for encoding and decoding strings.
         * @return DBR_STS_STRING, DBR_GR_STRING, or DBR_CTRL_STRING instance
         *         created from the serialized data read from
         *         <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessAlarmStringImpl deserialize(
                ChannelAccessValueType type, ByteSource byteSource, int count,
                Charset charset) {
            ChannelAccessAlarmStringImpl value = new ChannelAccessAlarmStringImpl(
                    type, charset);
            value.deserializeAlarms(byteSource);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Implementation of DBR_TIME_STRING.
     */
    public static final class ChannelAccessTimeStringImpl extends
            ChannelAccessAlarmStringBase<ChannelAccessTimeStringImpl> implements
            ChannelAccessTimeString {

        /**
         * Seconds part of time-stamp.
         */
        protected int timeStampSeconds;

        /**
         * Nanoseconds part of time-stamp.
         */
        protected int timeStampNanoseconds;

        /**
         * Creates a DBR_TIME_STRING.
         * 
         * @param value
         *            string value.
         * @param alarmStatus
         *            alarm status.
         * @param alarmSeverity
         *            alarm severity.
         * @param timeStampSeconds
         *            seconds part of time-stamp.
         * @param timeStampNanoSeconds
         *            nanoseconds part of time-stamp.
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        public ChannelAccessTimeStringImpl(Collection<String> value,
                ChannelAccessAlarmStatus alarmStatus,
                ChannelAccessAlarmSeverity alarmSeverity, int timeStampSeconds,
                int timeStampNanoSeconds, Charset charset) {
            super(value, alarmStatus, alarmSeverity, charset);
            this.timeStampSeconds = timeStampSeconds;
            if (timeStampNanoSeconds < 0 || timeStampNanoSeconds > 999999999) {
                throw new IllegalArgumentException(
                        "Nanoseconds part of timestamp must be a number between 0 and 999999999.");
            }
            this.timeStampNanoseconds = timeStampNanoSeconds;
        }

        /**
         * Creates a read-only DBR_TIME_STRING. All read operations are
         * delegated to <code>delegate</code>.
         * 
         * @param delegate
         *            value to which read operations are delegated.
         */
        private ChannelAccessTimeStringImpl(ChannelAccessTimeStringImpl delegate) {
            super(delegate);
        }

        /**
         * Special constructor only intended for deserialization code.
         * 
         * @param charset
         *            encoding used for encoding/decoding strings.
         */
        private ChannelAccessTimeStringImpl(Charset charset) {
            super(charset);
        }

        @Override
        public ChannelAccessValueType getType() {
            return ChannelAccessValueType.DBR_TIME_STRING;
        }

        @Override
        public ChannelAccessTimeStamp getTimeStamp() {
            if (delegate != null) {
                return delegate.getTimeStamp();
            }
            return ChannelAccessValueFactory.createTimeStamp(timeStampSeconds,
                    timeStampNanoseconds);
        }

        @Override
        public int getTimeSeconds() {
            if (delegate != null) {
                return delegate.getTimeSeconds();
            }
            return timeStampSeconds;
        }

        @Override
        public int getTimeNanoseconds() {
            if (delegate != null) {
                return delegate.getTimeNanoseconds();
            }
            return timeStampNanoseconds;
        }

        @Override
        public void setTimeStamp(ChannelAccessTimeStamp timeStamp) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.timeStampSeconds = timeStamp.getSeconds();
            this.timeStampNanoseconds = timeStamp.getNanoseconds();
        }

        @Override
        public void setTimeSeconds(int timeStampSeconds) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            this.timeStampSeconds = timeStampSeconds;
        }

        @Override
        public void setTimeNanoseconds(int timeStampNanoseconds) {
            if (delegate != null) {
                throw new UnsupportedOperationException(
                        "Cannot modify a read-only value.");
            }
            if (timeStampNanoseconds < 0 || timeStampNanoseconds > 999999999) {
                throw new IllegalArgumentException(
                        "Nanoseconds part of time-stamp must be a number between 0 and 999999999.");
            }
            this.timeStampNanoseconds = timeStampNanoseconds;
        }

        @Override
        public void serialize(ByteSink byteSink, int count) {
            assert (count >= 0);
            assert (count <= value.size());
            serializeAlarms(byteSink);
            ValueCodecUtils.serializeTime(byteSink, timeStampSeconds,
                    timeStampNanoseconds);
            serializeValue(byteSink, count);
        }

        @Override
        public ChannelAccessTimeString asReadOnlyValue() {
            if (delegate != null) {
                return this;
            }
            return new ChannelAccessTimeStringImpl(this);
        }

        @Override
        public boolean equals(Object obj) {
            if (delegate != null) {
                return delegate.equals(obj);
            }
            obj = extractDelegate(obj);
            if (!super.equals(obj)) {
                return false;
            }
            // Now, we can safely cast because the type has already been
            // checked.
            ChannelAccessTimeStringImpl other = (ChannelAccessTimeStringImpl) obj;
            return new EqualsBuilder()
                    .append(this.timeStampSeconds, other.timeStampSeconds)
                    .append(this.timeStampNanoseconds,
                            other.timeStampNanoseconds).isEquals();
        }

        @Override
        public int hashCode() {
            if (delegate != null) {
                return delegate.hashCode();
            }
            return new HashCodeBuilder().appendSuper(super.hashCode())
                    .append(this.timeStampSeconds)
                    .append(this.timeStampNanoseconds).toHashCode();
        }

        @Override
        public ChannelAccessTimeString clone() {
            if (delegate != null) {
                return delegate.clone();
            }
            return (ChannelAccessTimeString) super.clone();
        }

        /**
         * Creates a DBR_TIME_STRING from its serialized representation.
         * 
         * @param byteSource
         *            byte source to read data from.
         * @param count
         *            number of value elements to read.
         * @param charset
         *            character-set used for encoding and decoding strings.
         * @return DBR_TIME_STRING instance created from the serialized data
         *         read from <code>byteSource</code>.
         * @throws BufferUnderflowException
         *             if the byte source throws such an exception.
         */
        public static ChannelAccessTimeStringImpl deserialize(
                ByteSource byteSource, int count, Charset charset) {
            ChannelAccessTimeStringImpl value = new ChannelAccessTimeStringImpl(
                    charset);
            value.deserializeAlarms(byteSource);
            // Deserialize time-stamp.
            value.timeStampSeconds = ValueCodecUtils
                    .deserializeTimeSeconds(byteSource);
            value.timeStampNanoseconds = ValueCodecUtils
                    .deserializeTimeNanoseconds(byteSource);
            value.deserializeValue(byteSource, count);
            return value;
        }

    }

    /**
     * Base constructor.
     * 
     * @param value
     *            string value.
     * @param charset
     *            encoding used for encoding/decoding strings.
     */
    private ChannelAccessStringBase(Collection<String> value, Charset charset) {
        super(value, charset);
    }

    /**
     * Base constructor for read-only variant.
     * 
     * @param delegate
     *            value to which read operations are delegated.
     */
    private ChannelAccessStringBase(ImplementationType delegate) {
        super(delegate);
    }

    /**
     * Special constructor only intended for deserialization code.
     * 
     * @param charset
     *            encoding used for encoding/decoding strings.
     */
    private ChannelAccessStringBase(Charset charset) {
        super(charset);
    }

    @Override
    public ChannelAccessString clone() {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        @SuppressWarnings("unchecked")
        ChannelAccessStringBase<ImplementationType> copy = (ChannelAccessStringBase<ImplementationType>) super
                .clone();
        return copy;
    }

}
