/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import static org.junit.Assert.*;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.util.Arrays;

import org.junit.Test;

import com.aquenos.epics.jackie.common.io.ByteBufferByteSink;
import com.aquenos.epics.jackie.common.io.ByteBufferByteSource;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmSeverity;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueFactory;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Tests for the {@link ChannelAccessMessageCodec}.
 * 
 * @author Sebastian Marsching
 */
public class ChannelAccessMessageCodecTest {

    private static final Charset UTF8 = Charset.forName("UTF8");
    private static final Inet4Address LOCALHOST;
    static {
        try {
            LOCALHOST = (Inet4Address) InetAddress.getByName("127.0.0.1");
        } catch (UnknownHostException e) {
            throw new RuntimeException(
                    "Could not create IP address 127.0.0.1.", e);
        }
    }

    private ChannelAccessMessageCodec codec = new ChannelAccessMessageCodec();
    private ChannelAccessVersion version = ChannelAccessVersion.V4_13;

    private static ByteBufferByteSink createByteSink() {
        ByteBufferByteSink byteSink = new ByteBufferByteSink();
        byteSink.addBuffer(ByteBuffer.allocate(8192));
        return byteSink;
    }

    private static ByteBufferByteSource createByteSource(
            ByteBufferByteSink byteSink) {
        ByteBufferByteSource byteSource = new ByteBufferByteSource();
        byteSource.appendBuffers(byteSink.getWrittenData());
        return byteSource;
    }

    private ChannelAccessMessage reserializeMessageFromRepeater(
            ChannelAccessMessage message) {
        ByteBufferByteSink byteSink = createByteSink();
        codec.encodeMessageFromRepeater(byteSink, message, version);
        ByteBufferByteSource byteSource = createByteSource(byteSink);
        assertEquals(0, byteSource.remaining()
                % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT);
        ChannelAccessMessage newMessage = codec.decodeMessageFromRepeater(
                byteSource, version);
        assertEquals(0, byteSource.remaining());
        return newMessage;
    }

    private ChannelAccessMessage reserializeMessageToRepeater(
            ChannelAccessMessage message) {
        ByteBufferByteSink byteSink = createByteSink();
        codec.encodeMessageToRepeater(byteSink, message, version);
        ByteBufferByteSource byteSource = createByteSource(byteSink);
        assertEquals(0, byteSource.remaining()
                % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT);
        ChannelAccessMessage newMessage = codec.decodeMessageToRepeater(
                byteSource, version, LOCALHOST);
        assertEquals(0, byteSource.remaining());
        return newMessage;
    }

    private ChannelAccessMessage reserializeMessageToTCPClient(
            ChannelAccessMessage message) {
        ByteBufferByteSink byteSink = createByteSink();
        codec.encodeMessageToTCPClient(byteSink, message, version);
        ByteBufferByteSource byteSource = createByteSource(byteSink);
        assertEquals(0, byteSource.remaining()
                % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT);
        ChannelAccessMessage newMessage = codec.decodeMessageToTCPClient(
                byteSource, version, LOCALHOST, 1234);
        assertEquals(0, byteSource.remaining());
        return newMessage;
    }

    private ChannelAccessMessage reserializeMessageToTCPServer(
            ChannelAccessMessage message) {
        ByteBufferByteSink byteSink = createByteSink();
        codec.encodeMessageToTCPServer(byteSink, message, version);
        ByteBufferByteSource byteSource = createByteSource(byteSink);
        assertEquals(0, byteSource.remaining()
                % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT);
        ChannelAccessMessage newMessage = codec.decodeMessageToTCPServer(
                byteSource, version);
        assertEquals(0, byteSource.remaining());
        return newMessage;
    }

    private ChannelAccessMessage reserializeMessageToUDPClient(
            ChannelAccessMessage message) {
        ByteBufferByteSink byteSink = createByteSink();
        codec.encodeMessageToUDPClient(byteSink, message, version);
        ByteBufferByteSource byteSource = createByteSource(byteSink);
        assertEquals(0, byteSource.remaining()
                % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT);
        ChannelAccessMessage newMessage = codec.decodeMessageToUDPClient(
                byteSource, version, LOCALHOST, 1234);
        assertEquals(0, byteSource.remaining());
        return newMessage;
    }

    private ChannelAccessMessage reserializeMessageToUDPServer(
            ChannelAccessMessage message) {
        ByteBufferByteSink byteSink = createByteSink();
        codec.encodeMessageToUDPServer(byteSink, message, version);
        ByteBufferByteSource byteSource = createByteSource(byteSink);
        assertEquals(0, byteSource.remaining()
                % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT);
        ChannelAccessMessage newMessage = codec.decodeMessageToUDPServer(
                byteSource, version);
        assertEquals(0, byteSource.remaining());
        return newMessage;
    }

    /**
     * Encode and decode a REPEATER_CONFIRM message sent by a repeater.
     */
    @Test
    public void repeaterSendRepeaterConfirm() {
        ChannelAccessRepeaterConfirmMessage message = new ChannelAccessRepeaterConfirmMessage(
                LOCALHOST);
        assertEquals(message, reserializeMessageFromRepeater(message));
    }

    /**
     * Encode and decode a CA_PROTO_RSRV_IS_UP message sent by a repeater.
     */
    @Test
    public void repeaterSendBeacon() {
        ChannelAccessBeaconMessage message = new ChannelAccessBeaconMessage(
                LOCALHOST, 1234, version, 42);
        assertEquals(message, reserializeMessageFromRepeater(message));
    }

    /**
     * Encode and decode a CA_PROTO_RSRV_IS_UP message received by a repeater.
     */
    @Test
    public void repeaterReceiveBeacon() {
        ChannelAccessBeaconMessage message = new ChannelAccessBeaconMessage(
                LOCALHOST, 1234, version, 42);
        assertEquals(message, reserializeMessageToRepeater(message));
    }

    /**
     * Encode and decode a CA_PROTO_VERSION message sent by a TCP client.
     */
    @Test
    public void tcpClientVersion() {
        ChannelAccessVersionTCPClientMessage message = new ChannelAccessVersionTCPClientMessage(
                version, (short) 22);
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_EVENT_ADD message sent by a TCP client.
     */
    @Test
    public void tcpClientSubscription() {
        ChannelAccessSubscriptionClientMessage message = new ChannelAccessSubscriptionClientMessage(
                ChannelAccessValueType.DBR_CTRL_DOUBLE, 42, -123, 23,
                ChannelAccessEventMask.DBE_VALUE
                        .or(ChannelAccessEventMask.DBE_ALARM));
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_EVENT_CANCEL message sent by a TCP client.
     */
    @Test
    public void tcpClientCancelSubscription() {
        ChannelAccessCancelSubscriptionClientMessage message = new ChannelAccessCancelSubscriptionClientMessage(
                ChannelAccessValueType.DBR_STSACK_STRING, 55, 123, 42);
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_READ message sent by a TCP client.
     */
    @Test
    public void tcpClientRead() {
        ChannelAccessReadClientMessage message = new ChannelAccessReadClientMessage(
                ChannelAccessValueType.DBR_CLASS_NAME, 1, 5, 123);
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_WRITE message sent by a TCP client.
     */
    @Test
    public void tcpClientWrite() {
        ChannelAccessWriteMessage message = new ChannelAccessWriteMessage(
                ChannelAccessValueType.DBR_PUT_ACKS,
                1,
                -1,
                0,
                ChannelAccessValueFactory
                        .createAcknowledgeAlarm(ChannelAccessAlarmSeverity.MAJOR_ALARM));
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_EVENTS_OFF message sent by a TCP client.
     */
    @Test
    public void tcpClientEventsOff() {
        ChannelAccessEventsOffMessage message = new ChannelAccessEventsOffMessage();
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_EVENTS_ON message sent by a TCP client.
     */
    @Test
    public void tcpClientEventsOn() {
        ChannelAccessEventsOnMessage message = new ChannelAccessEventsOnMessage();
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_READ_SYNC message sent by a TCP client.
     */
    @Test
    public void tcpClientReadSync() {
        ChannelAccessReadSyncMessage message = new ChannelAccessReadSyncMessage();
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_CLEAR_CHANNEL message sent by a TCP client.
     */
    @Test
    public void tcpClientDisconnectChannel() {
        ChannelAccessDisconnectChannelMessage message = new ChannelAccessDisconnectChannelMessage(
                128, 821);
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_READ_NOTIFY message sent by a TCP client.
     */
    @Test
    public void tcpClientReadNotify() {
        ChannelAccessReadNotifyClientMessage message = new ChannelAccessReadNotifyClientMessage(
                ChannelAccessValueType.DBR_CLASS_NAME, 1, 5, 123);
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_CREATE_CHAN message sent by a TCP client.
     */
    @Test
    public void tcpClientConnectChannel() {
        ChannelAccessConnectChannelClientMessage message = new ChannelAccessConnectChannelClientMessage(
                42, version, "myChannel");
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_WRITE_NOTIFY message sent by a TCP client.
     */
    @Test
    public void tcpClientWriteNotify() {
        ChannelAccessWriteNotifyClientMessage message = new ChannelAccessWriteNotifyClientMessage(
                ChannelAccessValueType.DBR_DOUBLE, 4, -1, 0,
                ChannelAccessValueFactory.createDouble(new double[] { 24, 8,
                        13, 5 }));
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_CLIENT_NAME message sent by a TCP client.
     */
    @Test
    public void tcpClientUserName() {
        ChannelAccessUserNameMessage message = new ChannelAccessUserNameMessage(
                "user42");
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_HOST_NAME message sent by a TCP client.
     */
    @Test
    public void tcpClientHostName() {
        ChannelAccessHostNameMessage message = new ChannelAccessHostNameMessage(
                "www.example.com");
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_ECHO message sent by a TCP client.
     */
    @Test
    public void tcpClientEcho() {
        ChannelAccessEchoMessage message = new ChannelAccessEchoMessage();
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_ECHO message sent by a TCP client.
     */
    @Test
    public void tcpClientSearch() {
        ChannelAccessSearchClientMessage message = new ChannelAccessSearchClientMessage(
                555, version, false, "myChannel");
        assertEquals(message, reserializeMessageToTCPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_VERSION message sent by a TCP server.
     */
    @Test
    public void tcpServerVersion() {
        ChannelAccessVersionTCPServerMessage message = new ChannelAccessVersionTCPServerMessage(
                version);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_EVENT_ADD message sent by a TCP server.
     */
    @Test
    public void tcpServerSubscription() {
        ChannelAccessSubscriptionServerMessage message = new ChannelAccessSubscriptionServerMessage(
                ChannelAccessValueType.DBR_STRING, 1,
                ChannelAccessStatus.StatusMessage.ECA_NORMAL.toStatus(), 123,
                ChannelAccessValueFactory.createString(
                        Arrays.asList(new String[] { "abc" }), UTF8));
        assertEquals(message, reserializeMessageToTCPClient(message));
        message = new ChannelAccessSubscriptionServerMessage(
                ChannelAccessValueType.DBR_STRING, 1,
                ChannelAccessStatus.StatusMessage.ECA_ALLOCMEM.toStatus(), 123,
                null);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_EVENT_ADD message sent by a TCP server.
     */
    @Test
    public void tcpServerCancelSubscription() {
        ChannelAccessCancelSubscriptionServerMessage message = new ChannelAccessCancelSubscriptionServerMessage(
                ChannelAccessValueType.DBR_STRING, 1, 453, 123);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_READ message sent by a TCP server.
     */
    @Test
    public void tcpServerRead() {
        ChannelAccessReadServerMessage message = new ChannelAccessReadServerMessage(
                ChannelAccessValueType.DBR_STS_FLOAT, 1, 0, 99,
                ChannelAccessValueFactory.createAlarmFloat(
                        new float[] { 3.3f },
                        ChannelAccessAlarmStatus.NO_ALARM,
                        ChannelAccessAlarmSeverity.NO_ALARM));
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_READ_SYNC message sent by a TCP server.
     */
    @Test
    public void tcpServerReadSync() {
        ChannelAccessReadSyncMessage clientMessage = new ChannelAccessReadSyncMessage();
        ChannelAccessReadSyncMessage message = new ChannelAccessReadSyncMessage(
                clientMessage);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_ERROR message sent by a TCP server.
     */
    @Test
    public void tcpServerError() {
        // We have to use a message without a payload here, because otherwise
        // the message embedded in the error message will lose its payload and
        // thus the error message sent will not be equal to the error message
        // received.
        ChannelAccessReadNotifyClientMessage clientMessage = new ChannelAccessReadNotifyClientMessage(
                ChannelAccessValueType.DBR_CTRL_DOUBLE, 5, 123, 999);
        // We have to reserialize the client message so that the original
        // headers are available.
        clientMessage = (ChannelAccessReadNotifyClientMessage) reserializeMessageToTCPServer(clientMessage);
        ChannelAccessErrorMessage message = new ChannelAccessErrorMessage(99,
                ChannelAccessStatus.StatusMessage.ECA_ALLOCMEM.toStatus(),
                clientMessage, "Cannot allocate memory.");
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_CLEAR_CHANNEL message sent by a TCP server.
     */
    @Test
    public void tcpServerDisconnectChannel() {
        ChannelAccessDisconnectChannelMessage message = new ChannelAccessDisconnectChannelMessage(
                99, 42);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_READ_NOTIFY message sent by a TCP server.
     */
    @Test
    public void tcpServerReadNotify() {
        ChannelAccessReadNotifyServerMessage message = new ChannelAccessReadNotifyServerMessage(
                ChannelAccessValueType.DBR_STS_LONG, 1,
                ChannelAccessStatus.StatusMessage.ECA_NORMAL.toStatus(), 99,
                ChannelAccessValueFactory.createAlarmLong(new int[] { 33 },
                        ChannelAccessAlarmStatus.NO_ALARM,
                        ChannelAccessAlarmSeverity.NO_ALARM));
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_CREATE_CHAN message sent by a TCP server.
     */
    @Test
    public void tcpServerConnectChannel() {
        ChannelAccessConnectChannelServerMessage message = new ChannelAccessConnectChannelServerMessage(
                55, 23, ChannelAccessValueType.DBR_ENUM, 20);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_WRITE_NOTIFY message sent by a TCP server.
     */
    @Test
    public void tcpServerWriteNotify() {
        ChannelAccessWriteNotifyServerMessage message = new ChannelAccessWriteNotifyServerMessage(
                ChannelAccessValueType.DBR_SHORT, 1,
                ChannelAccessStatus.StatusMessage.ECA_NORMAL.toStatus(), 22);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_ACCESS_RIGHTS message sent by a TCP server.
     */
    @Test
    public void tcpServerAccessRights() {
        ChannelAccessAccessRightsMessage message = new ChannelAccessAccessRightsMessage(
                321, true, false);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_ECHO message sent by a TCP server.
     */
    @Test
    public void tcpServerEcho() {
        ChannelAccessEchoMessage message = new ChannelAccessEchoMessage();
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_CREATE_CH_FAIL message sent by a TCP server.
     */
    @Test
    public void tcpServerConnectChannelFailed() {
        ChannelAccessConnectChannelFailedMessage message = new ChannelAccessConnectChannelFailedMessage(
                234);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_SERVER_DISCONN message sent by a TCP server.
     */
    @Test
    public void tcpServerChannelDisconnectedByServer() {
        ChannelAccessChannelDisconnectedByServerMessage message = new ChannelAccessChannelDisconnectedByServerMessage(
                -789);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_SEARCH message sent by a TCP server.
     */
    @Test
    public void tcpServerSearchTCPServer() {
        ChannelAccessSearchTCPServerMessage message = new ChannelAccessSearchTCPServerMessage(
                987, LOCALHOST, 1234);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_NOT_FOUND message sent by a TCP server.
     */
    @Test
    public void tcpServerNotFound() {
        ChannelAccessSearchClientMessage clientMessage = new ChannelAccessSearchClientMessage(
                999, version, true, "myChannel");
        ChannelAccessNotFoundMessage message = new ChannelAccessNotFoundMessage(
                clientMessage);
        assertEquals(message, reserializeMessageToTCPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_VERSION message sent by a UDP client.
     */
    @Test
    public void udpClientVersion() {
        ChannelAccessVersionUDPMessage message = new ChannelAccessVersionUDPMessage(
                version, true, 25);
        assertEquals(message, reserializeMessageToUDPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_SEARCH message sent by a UDP client.
     */
    @Test
    public void udpClientSearch() {
        ChannelAccessSearchClientMessage message = new ChannelAccessSearchClientMessage(
                456, version, false, "someChannel");
        assertEquals(message, reserializeMessageToUDPServer(message));
    }

    /**
     * Encode and decode a CA_PROTO_VERSION message sent by a UDP server.
     */
    @Test
    public void udpServerVersion() {
        ChannelAccessVersionUDPMessage message = new ChannelAccessVersionUDPMessage(
                version, false, 37);
        assertEquals(message, reserializeMessageToUDPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_SEARCH message sent by a UDP server.
     */
    @Test
    public void udpServerSearch() {
        ChannelAccessSearchUDPServerMessage message = new ChannelAccessSearchUDPServerMessage(
                version, 321, LOCALHOST, 1234);
        assertEquals(message, reserializeMessageToUDPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_ERROR message sent by a UDP server.
     */
    @Test
    public void udpServerError() {
        ChannelAccessErrorMessage message = new ChannelAccessErrorMessage(4321,
                ChannelAccessStatus.StatusMessage.ECA_BADSTR.toStatus(), null,
                null);
        assertEquals(message, reserializeMessageToUDPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_NOT_FOUND message sent by a UDP server.
     */
    @Test
    public void udpServerNotFound() {
        ChannelAccessSearchClientMessage clientMessage = new ChannelAccessSearchClientMessage(
                4321, version, true, "abcdef");
        ChannelAccessNotFoundMessage message = new ChannelAccessNotFoundMessage(
                clientMessage);
        assertEquals(message, reserializeMessageToUDPClient(message));
    }

    /**
     * Encode and decode a CA_PROTO_ECHO message sent by a UDP server.
     */
    @Test
    public void udpServerEcho() {
        ChannelAccessEchoMessage message = new ChannelAccessEchoMessage();
        assertEquals(message, reserializeMessageToUDPClient(message));
    }

}
